import rkgb
import torch
import numpy as np
from copy import deepcopy, copy

from rkgb.lowlevel.ast_add_on import ast_to_str
from rkgb.core.hierarchical import HierarchicalGraph, HierarchicalCluster
from rkgb.core.backward import ComputationNode
from rkgb.lowlevel.constants import init_target_string

from ..op_schedule import (
    OpSchedule,
    ComputeOp,
    DeleteOp,
    Activation,
    OffloadOp,
    PrefetchOp,
    AllocateOp,
)
import time
import psutil


class Solver:
    class Config:
        def __init__(self):
            pass

    def __init__(self, config=None, **kwargs):
        self.config = config if config is not None else type(self).Config(**kwargs)

    def __call__(self, cluster: HierarchicalCluster, budgets=None, *args, **kargs):
        return self.solve(cluster, budgets, *args, **kargs)

    def solve(self, cluster: HierarchicalCluster, budgets=None):
        # -> RETURN list of Op_sched
        pass


class FastSolver(Solver):
    def __init__(self, config=None, recompute_sched=True):
        super().__init__(config)
        self.recompute_sched=recompute_sched

    def solve_hcn(self, hcn, cluster, no_del_names):
        if hcn.sub_cluster is None:  # fwd with no grad
            if hcn.name in cluster.dict_nodes:
                cnode = cluster.dict_nodes[hcn.name]
                hcn.ff_time = cnode.time
                hcn.ff_overhead = cnode.mem_overhead
                hcn.ff_op_list = [ComputeOp(cnode, fast_forward=True)]
            else:  # loss node
                hcn.ff_time = 0
                hcn.ff_overhead = 0
                hcn.ff_op_list = []
            return

        re_cluster = hcn.sub_cluster.representee_cluster
        list_sched = self.solve(re_cluster)

        if re_cluster.list_schedules == []:
            # re_cluster.list_schedules += self.solve(re_cluster)
            re_cluster.list_schedules.append(list_sched[0])
            if self.recompute_sched and not re_cluster.is_bottom:
                re_cluster.list_schedules.append(list_sched[1])
        autograd_sched = list_sched[1]
        hcn.ff_time = autograd_sched.fwd_time
        hcn.ff_overhead = autograd_sched.fwd_overhead
        hcn.ff_op_list = translate(
            hcn.sub_cluster, autograd_sched.op_list[: autograd_sched.loss_idx]
        )

    def solve(
        self,
        cluster: HierarchicalCluster,
        no_del_names=[f"{init_target_string} data", f"{init_target_string} grad"],
        recompute_sched=False,
    ):
        """
        Return basic schedules for the cluster:
        1. PyTorch autograd schedule.
        2. Save_none: forward saves no intermediate activations and delete ASAP;
                      backward runs autograd schedule (fwd+bwd).
        """
        assert cluster is cluster.representee_cluster

        list_sched = []
        # if not cluster is cluster.representee_cluster:
        #     return list_sched
        ff_op_list = self.single_compute_op_list(
            cluster,
            with_backward=False,
            no_del_names=no_del_names,
            fast_forward=True,
        )
        # list_sched.append(
        #     OpSchedule(
        #     ff_op_list + [ComputeOp(ComputationNode("loss"))],
        #     cluster=cluster,
        #     loss_idx=len(ff_op_list)
        #     )
        # )  # not real sched, only for info

        autograd_op_list = self.single_compute_op_list(
            cluster,
            no_del_names=no_del_names,
        )
        autograd_loss_idx = autograd_op_list.index(
            ComputeOp(ComputationNode("loss"), disabled=True)
        )
        autograd_sched = OpSchedule(
            autograd_op_list, cluster=cluster, loss_idx=autograd_loss_idx
        )
        re_autograd_op_list = copy(autograd_op_list) ##LED was deepcopy()
        loss_op = re_autograd_op_list.pop(autograd_loss_idx)
        recompute_op_list = ff_op_list + [loss_op] + re_autograd_op_list
        list_sched.append(autograd_sched)

        list_sched.append(
            OpSchedule(recompute_op_list, cluster=cluster, loss_idx=len(ff_op_list))
        )
        return list_sched

    def single_compute_op_list(
        self,
        cluster: HierarchicalCluster,
        with_backward=True,
        no_del_names=[],
        fast_forward=False,
    ):
        list_cnode = cluster.list_cnodes.copy()
        if not with_backward:
            list_cnode = list_cnode[: cluster.loss_idx]
        loss_i = (
            list_cnode.index(cluster.loss_cnode)
            if cluster.loss_cnode in list_cnode
            else -1
        )

        def _can_del(i, anode):
            if anode.name in no_del_names:
                return False
            for cnode in anode.users_real:  # .union(anode.users_fake):
                if cnode in list_cnode[i + 1 :]:
                    return False

            fwd = not with_backward or i <= loss_i
            if anode in cluster.loss_cnode.deps_real and fwd:
                return False
            # if anode in cluster.interfaces["input_data_anodes"]:
            #     return False
            if anode in cluster.interfaces["input_grad_anodes"]:
                return False
            return True

        op_list = []
        alive_status = {}

        for anode in cluster.list_anodes:
            # if hdn not in cluster.interfaces:
            alive_status[anode.name] = (
                1 if (anode in cluster.interfaces["input_data_anodes"]) else 0
            )

        loss_idx = None
        for i, cnode in enumerate(list_cnode):
            for anode in cnode.users:
                if "phantom" in anode.name and fast_forward:
                    continue
                alive_status[anode.name] = 1
            if cnode == cluster.loss_cnode:
                loss_idx = len(op_list)
            op_list.append(
                ComputeOp(
                    cnode,
                    detach=True,  # not with_bwd,
                    fast_forward=fast_forward,
                    disabled=("loss" in cnode.name),
                )
            )

            for anode_name, alive in alive_status.items():
                if alive:
                    anode = cluster.dict_nodes[anode_name]
                    if _can_del(i, anode):
                        op_list.append(DeleteOp(Activation(anode)))
                        alive_status[anode_name] = 0

        return op_list

    # def recursive_solve_hcn(self,
    #                         hcn,
    #                         no_del_names=[f"{init_target_string} data",
    #                         f"{init_target_string} grad"]):
    #     cluster = hcn.sub_cluster
    #     if not cluster is cluster.representee_cluster or cluster.is_bottom:
    #         return
    #     for hg in cluster.partitionings:
    #         for hcn in hg.list_HCNs:
    #             self.recursive_solve_hcn(hcn, hcn, no_del_names)
    #             self.solve_hcn(hcn, no_del_names)

    def preprocess(
        self,
        cluster: HierarchicalCluster,
        no_del_names=[f"{init_target_string} data", f"{init_target_string} grad"],
    ):
        if not cluster is cluster.representee_cluster or cluster.is_bottom:
            return
        for hg in cluster.partitionings:
            for hcn in hg.list_HCNs:
                self.solve_hcn(hcn, cluster, no_del_names)


def add_sched(cluster: HierarchicalCluster, sched):
    cluster.list_schedules.append(sched)


def get_sched(cluster: HierarchicalCluster, pareto=False):
    representee = cluster.representee_cluster
    if not pareto:
        return representee.list_schedules
    else:
        list_schedules = representee.list_schedules
        time_mem = np.array(
            [(sum(op_sched.time), op_sched.mem) for op_sched in list_schedules]
        )
        is_pareto = np.ones(len(list_schedules), dtype=bool)
        for i, c in enumerate(time_mem):
            is_pareto[i] = np.all(np.any(time_mem >= c, axis=1))

    return [list_schedules[i] for i, p in enumerate(is_pareto) if p]


def translate(cluster: HierarchicalCluster, op_list):
    if cluster is cluster.representee_cluster:
        return op_list
    translator_re = cluster.representee_cluster.translator
    translator = cluster.translator
    translated_op_list = [ copy(op) for op in op_list]

    def translate_op(op):
        if isinstance(op, ComputeOp):
            ana_kn = translator_re.to_ano(op.target)
            op.target = translator.from_ano(ana_kn)
        elif isinstance(op.target, Activation):
            ana_kn = translator_re.to_ano(op.target.anode)
            op.target = Activation(translator.from_ano(ana_kn))
        else:
            raise ValueError

    for op in translated_op_list:
        translate_op(op)
        for e in op.wait_events:
            re_node = translator_re.to_ano(e[1])
            e[1] = translator.from_ano(re_node)
    return translated_op_list


def get_hgraph_budget_lb(hgraph: HierarchicalGraph):
    # Lower bound for minimum feasible budget given schedules
    hcn_memory_budget = []
    for hcn in hgraph.list_HCNs:
        if hcn.sub_cluster is not None:
            # list_schedules = hcn.sub_cluster.get_sched()
            list_schedules = get_sched(hcn.sub_cluster)
            hcn_memory_budget.append(
                min(op_sched.peak_mem for op_sched in list_schedules)
            )
        else:
            hcn_memory_budget.append(hcn.ff_overhead)
    return max(hcn_memory_budget, default=0)


def get_hgraph_budget_ub(hgraph: HierarchicalGraph):
    # Upper bound for minimum feasible budget given schedules
    cluster = hgraph.cluster
    if cluster.representee_cluster.list_schedules == []:
        autograd_op_list = get_single_compute_op_list(
            cluster,
            with_bwd=True,
        )
        autograd_sched = OpSchedule(
            autograd_op_list,
            cluster=cluster,
        )
    else:
        autograd_sched = cluster.representee_cluster.list_schedules[0]
    max_bdg = autograd_sched.mem + autograd_sched.bwd_overhead
    return max_bdg


def get_cluster_budget(
    cluster: HierarchicalCluster,
    nb_bdg_peak=3,
    nb_bdg_save=6,
    overall_bdg=None,
    with_save_budget=False,
):
    # assuming solving budget does not based on lower level solution
    budgets = []

    sizes = [anode.mem for anode in cluster.list_anodes]
    overheads = [
        cnode.mem_overhead for cnode in cluster.list_cnodes if cnode.mem_overhead is not None
    ]

    # overheads = [hcn.sub_cluster.ff_overhead for hcn in hg.list_HCNs] + [
    #     op_sched.bwd_overhead for op_sched in hg.list_schedules
    # ]
    # max_bdg = sum(sizes) + max(overheads)
    if cluster.representee_cluster.list_schedules == []:
        autograd_op_list = get_single_compute_op_list(
            cluster,
            with_bwd=True,
        )
        autograd_sched = OpSchedule(
            autograd_op_list,
            cluster=cluster,
        )
    else:
        autograd_sched = cluster.representee_cluster.list_schedules[0]
    interfaces_mem = sum(anode.mem for anode in cluster.all_interfaces)
    max_bdg = autograd_sched.mem + autograd_sched.bwd_overhead
    if overall_bdg is not None:
        max_bdg = min(max_bdg, overall_bdg)
    min_bdg = max(overheads)
    # max_bdg = hg.list_schedules[0].mem + max(overheads)

    # TODO: find the minimum feasible budget
    # min_bdg = hg.fast_fwd_overhead()[0]
    # min_bdg = min(op_sched.mem for op_sched in hg.list_schedules) + max(overheads)

    l_bd_peak = np.linspace(min_bdg, max_bdg, nb_bdg_peak) + interfaces_mem
    if not with_save_budget:
        return l_bd_peak
    for bd_peak in l_bd_peak:
        l_bd_save = (
            np.linspace(
                0,
                min(bd_peak, autograd_sched.mem),
                nb_bdg_save,
            )
            + interfaces_mem
        )
        # for bd_save in l_bd_save:
        #     budgets.append((bd_peak, bd_save))
        budgets.append((bd_peak, l_bd_save))
    return budgets


def solve_recursive(h_cluster: HierarchicalCluster, list_solvers=[], skip_self=False):
    # assume it's representee
    # print(h_cluster.name)
    for hg in h_cluster.partitionings:
        # print(hg.name)
        for hcn in hg.list_HCNs:
            if (
                hcn.is_fwd
                and hcn.sub_cluster is not None
                and not hcn.sub_cluster.is_bottom
                and hcn.sub_cluster is hcn.sub_cluster.representee_cluster
            ):
                sub_cluster = hcn.sub_cluster
                solve_recursive(sub_cluster, list_solvers)
    if not skip_self:
        for solver in list_solvers:
            # h_cluster.solve(solver)
            if h_cluster is h_cluster.representee_cluster:
                last_time = time.time()
                h_cluster.list_schedules.extend(solver(h_cluster))


# Preprocessing Cluster: add fast_forward and autograd option
def preprocess_rec(cluster: HierarchicalCluster):
    if cluster is cluster.representee_cluster:
        if not cluster.is_bottom:
            for hg in cluster.partitionings:
                for hcn in hg.list_HCNs:
                    if hcn.is_fwd and hcn.sub_cluster is not None:
                        # if not hcn.sub_cluster.list_schedules:
                        preprocess_rec(hcn.sub_cluster)
            preprocess(cluster)


def preprocess(
    cluster: HierarchicalCluster,
    no_del_names=[f"{init_target_string} data", f"{init_target_string} grad"],
    add_no_save_sched=True,
):
    if cluster is cluster.representee_cluster:
        for hg in cluster.partitionings:
            for hcn in hg.list_HCNs:
                if hcn.sub_cluster is None:  # fwd with no grad
                    if hcn.name in cluster.dict_nodes:
                        cnode = cluster.dict_nodes[hcn.name]
                        ff_op_list = [ComputeOp(cnode, fast_forward=True)]
                        hcn.ff_time = cnode.time
                        hcn.ff_overhead = cnode.mem_overhead
                    else:  # loss node
                        ff_op_list = []
                        hcn.ff_time = 0
                        hcn.ff_overhead = 0
                else:
                    ff_op_list = get_single_compute_op_list(
                        hcn.sub_cluster,
                        with_bwd=False,
                        no_del_names=no_del_names,
                        ff=True,
                    )
                    ff_op_sched = OpSchedule(
                        ff_op_list + [ComputeOp(ComputationNode("loss"))],
                        cluster=cluster,
                        loss_idx=len(ff_op_list),
                        # correct_overhead=False,
                    )  # not real sched, only for info
                    hcn.ff_time = ff_op_sched.fwd_time
                    hcn.ff_overhead = ff_op_sched.fwd_overhead
                    if (
                        hcn.sub_cluster.representee_cluster is hcn.sub_cluster
                        and hcn.sub_cluster.list_schedules == []
                    ):
                        autograd_op_list = get_single_compute_op_list(
                            hcn.sub_cluster,
                            with_bwd=True,
                            no_del_names=no_del_names,
                        )
                        hcn.sub_cluster.list_schedules.append(
                            OpSchedule(
                                autograd_op_list,
                                cluster=hcn.sub_cluster,
                                loss_idx=autograd_op_list.index(
                                    ComputeOp(ComputationNode("loss"), disabled=True)
                                ),
                            )
                        )
                        if add_no_save_sched:
                            hcn.sub_cluster.list_schedules.append(
                                OpSchedule(
                                    ff_op_list
                                    + [ComputeOp(ComputationNode("loss"))]
                                    + autograd_op_list,
                                    cluster=hcn.sub_cluster,
                                    loss_idx=len(ff_op_list),
                                )
                            )
                hcn.ff_op_list = ff_op_list


def get_single_compute_op_list(
    cluster: HierarchicalCluster, with_bwd=True, no_del_names=[], ff=False
):
    list_cnode = cluster.list_cnodes.copy()
    if not with_bwd:
        list_cnode = list_cnode[: cluster.loss_idx]
    loss_i = (
        list_cnode.index(cluster.loss_cnode) if cluster.loss_cnode in list_cnode else -1
    )

    def _can_del(i, anode):
        if anode.name in no_del_names:
            return False
        # for cnode in list_cnode[i + 1 :]:
        #     if anode in cnode.deps_real:
        #         return False
        for cnode in anode.users_real:
            if cnode in list_cnode[i + 1 :]:
                return False

        if anode in cluster.loss_cnode.deps_real and i <= loss_i:
            return False
        # if anode in cluster.interfaces["input_data_anodes"]:
        #     return False
        if anode in cluster.interfaces["input_grad_anodes"]:
            return False
        return True

    op_list = []
    # alive_list = []
    alive_status = {}
    for anode in cluster.list_anodes:
        # if hdn not in cluster.interfaces:
        alive_status[anode.name] = (
            1 if (anode in cluster.interfaces["input_data_anodes"]) else 0
        )

    loss_idx = None
    for i, cnode in enumerate(list_cnode):
        for anode in cnode.users:
            if "phantom" in anode.name and ff:
                continue
            alive_status[anode.name] = 1
        if cnode == cluster.loss_cnode:
            loss_idx = len(op_list)
        op_list.append(
            ComputeOp(
                cnode,
                detach=True,  # not with_bwd,
                fast_forward=ff,
                disabled=("loss" in cnode.name),
            )
        )

        for anode_name, alive in alive_status.items():
            if alive:
                anode = cluster.dict_nodes[anode_name]
                if _can_del(i, anode):
                    op_list.append(DeleteOp(Activation(anode)))
                    alive_status[anode_name] = 0

    return op_list  # , loss_idx





from copy import deepcopy
import numpy as np
import torch

import rkgb
# from .def_op import RunOp, DelOp, OpSchedule

import subprocess
import os
import tempfile

HOMEDIR = os.path.expanduser("~")
TWREMAT = os.path.join(HOMEDIR, "twremat")


# Allow users to pass 'humanized' memlimit values as strings.
def parse_memlimit(memlimit):
    if memlimit[-1] == "K":
        return int(memlimit[:-1]) * 1024
    elif memlimit[-1] == "M":
        return int(memlimit[:-1]) * 1024**2
    elif memlimit[-1] == "G":
        return int(memlimit[:-1]) * 1024**3
    else:
        return int(memlimit)


def runtwremat(gr, memlimit, target, loss, verbose=False):
    if type(memlimit) is str:
        memlimit = parse_memlimit(memlimit)

    fname = tempfile.mktemp()
    outname = tempfile.mktemp()

    with open(fname, "w") as fp:
        print("p remat2", file=fp)

        if memlimit != None:
            print(f"memlimit {memlimit}", file=fp)

        for n, info in gr.items():
            deps = " ".join(str(d) for d in info["deps"])

            if info["type"] == "normal":
                cpu = info["cpu"]
                mem = info["mem"]
                overhead_value = info["overhead"]
                weight = f"cpu {cpu} mem {mem} overhead {overhead_value}"
            elif info["type"] == "effectful":
                weight = "effectful"
            elif info["type"] == "pointer":
                weight = "pointer"
            if n in target:
                tstr = "target"
            else:
                tstr = ""

            if n in loss:
                loss_str = "loss"
            else:
                loss_str = ""

            print(f"node {n} deps {deps} {weight} {tstr} {loss_str}", file=fp)

    if verbose:
        print(" ".join([TWREMAT, fname, outname]))
    result = subprocess.run(["twremat", fname, outname], check=True, capture_output=not verbose)

    out = []
    with open(outname, "r") as fp:
        for line in fp:
            line = line.split()
            if line and line[0] == "c":
                out.append(("compute", int(line[1])))
            elif line and line[0] == "f":
                out.append(("free", int(line[1])))
            elif line:
                print(line)
                exit()
    return out


def get_rockmate_graphs(model, sample, device="cuda"):
    _model = deepcopy(model).to(device)
    for _, p in _model.named_parameters():
        if p.grad is None:
            p.grad = torch.zeros_like(p)

    if isinstance(sample, list):
        _sample = []
        for input in sample:
            _sample.append(deepcopy(input).to(device))
    else:
        _sample = deepcopy(sample).to(device)

    # Get graphs with rk-GB
    try:
        rkgb_res = rkgb.rkgb.Result(model=_model,
                               model_args=_sample,
                               inspection_device=torch.device("cuda"))
        H_cluster = rkgb_res.hierarchical_cluster

        ### OLD RKGB START   

        # rkgb_res = rkgb.make_all_graphs(
        #    _model, _sample, check_device_is_gpu=True
        # )
        # # rkgb.print_all_graphs(rkgb_res,name="fno1d",render_format="pdf")
        # # forward_graph = rkgb_res.S_graph
        # K_graph = rkgb_res.K_graph

        # K_graph.fake_input_kdn_grad() ## FOR ORIGINAL ROCKMATE COMPATIBILITY

        ### OLD RKGB END

    except Exception as e:
        del _model
        del _sample
        print(e)
        raise Exception(
            "K-graph can't be built on one GPU for given model hyperparameters"
        )

    del _model
    del _sample

    #return K_graph
    return H_cluster


def get_twremat_graph(H_cluster, contains_data_node=False, allow_loss_recomputation=False
):
    """Builds fw-bckw graph, containing only computational nodes

    For each computational node `cnode` from K_graph collects:
    - memory info on input data tensors, `mem_inputs`
    - memory info on output data tensors, `mem_outputs`
    - memory info on computational overhead, `cnode.overhead`
    - set of parent computational nodes, `deps`

    Parameters
    ----------
    K_graph: rkgb.K_graph
        A data flow graph obrained with rk-GB, it has two types of nodes: data and computation nodes.

    Returns
    -------
        dict of dict
            Returns a dictionary `node_info`, where keys are ids of computational nodes and each value is a dictionary,
            which contains collected info on cnode in the format compatible with input to Haskell code of TWRemat, i.e.
        ::

            node_info[cnode.unique_id]= {
            'type': 'normal',
            'cpu': mem_inputs + mem_outputs,
            'mem': mem_outputs,
            'overhead': cnode.mem_overhead,
            'deps': cnode_deps
            }
        int
            Id of the computational node which is indicated as a final node (the node, which does not have any children nodes)
    """

    node_info = {}
    targets = []
    for cnode in H_cluster.list_cnodes:
        users_inside = [
            dnode
            for dnode in cnode.users
            if (dnode in H_cluster.list_anodes and not dnode in H_cluster.interfaces['input_grad_anodes'])
        ]
        if len(users_inside) == 0:
            targets.append(cnode.unique_id)

    for cnode in H_cluster.list_cnodes:
        mem_inputs = 0
        mem_outputs = 0
        cnode_deps = [
            cn.unique_id 
            for cn in cnode.deps_through_artifacts
            if cn in H_cluster.list_cnodes
        ]
        for dnode in cnode.deps_real:
            mem_inputs += dnode.mem

            for cn in H_cluster.list_cnodes:
                if dnode in cn.users:
                    cnode_deps.append(cn.unique_id)

        for dnode in cnode.deps_fake:
            mem_inputs += dnode.mem

            for cn in H_cluster.list_cnodes:
                if dnode in cn.users:
                    cnode_deps.append(cn.unique_id)

        # if (cnode.main_target != 'loss'):
        if not cnode.unique_id in targets:
            for dnode in cnode.users:
                mem_outputs += dnode.mem


        node_info[cnode.unique_id] = {
            "type": "normal",
            "cpu": mem_inputs + mem_outputs,
            "mem": mem_outputs,
            "overhead": cnode.mem_overhead if cnode.mem_overhead else 0,
            "deps": cnode_deps,
        }

    if not allow_loss_recomputation:
        loss_node = [
            [node for node in H_cluster.list_cnodes if "loss" in node.name][
                0
            ].unique_id
        ]
    else:
        loss_node = []

    if not targets:
        raise ValueError("no targets")
    if len(node_info) < 1:
        raise ValueError("no node_info")
    # if not loss_node:
    #     raise ValueError("no loss_node")
    return node_info, targets, loss_node

    # ### OLD RKGB START

    # # target = K_graph.list_kcn[-1].unique_id  # final computational node
    # targets = []
    # for kcn in K_graph.list_kcn:
    #     users_inside = [
    #         dn
    #         for dn in kcn.users
    #         if (dn in K_graph.list_kdn and not dn in K_graph.interfaces["inputs_kdn_grad"])
    #     ]
    #     if len(users_inside) == 0:
    #         targets.append(kcn.unique_id)

    # # for cnode in K_graph.list_kcn:
    # #     mem_inputs = 0
    # #     mem_outputs = 0
    # #     cnode_deps = []

    # #     for dnode in cnode.deps_real:
    # #         mem_inputs += dnode.mem

    # #         for cn in dnode.deps:
    # #             cnode_deps.append(cn.unique_id)

    # for cnode in K_graph.list_kcn:
    #     mem_inputs = 0
    #     mem_outputs = 0
    #     cnode_deps = [
    #         cn.unique_id
    #         for cn in cnode.deps_through_size_artifacts
    #         if cn in K_graph.list_kcn
    #     ]

    #     for dnode in cnode.deps_real:
    #         mem_inputs += dnode.mem

    #         for cn in K_graph.list_kcn:
    #             if dnode in cn.users:
    #                 cnode_deps.append(cn.unique_id)

    #     # for dnode in cnode.deps_fake:
    #     #     mem_inputs += dnode.mem
    #     #     for cn in dnode.deps:
    #     #         cnode_deps.append(cn.unique_id)

        # for dnode in cnode.deps_fake:
        #     mem_inputs += dnode.mem

        #     for cn in K_graph.list_kcn:
        #         if dnode in cn.users:
        #             cnode_deps.append(cn.unique_id)

        # # if (cnode.main_target != 'loss'):
        # if not cnode.unique_id in targets:
        #     for dnode in cnode.users:
        #         mem_outputs += dnode.mem

    #     node_info[cnode.unique_id] = {
    #         "type": "normal",
    #         "cpu": mem_inputs + mem_outputs,
    #         "mem": mem_outputs,
    #         "overhead": cnode.overhead if cnode.overhead else 0,
    #         "deps": cnode_deps,
    #     }

    # if not allow_loss_recomputation:
    #     loss_node = [
    #         [node for node in K_graph.list_kcn if "loss" in node.name][
    #             0
    #         ].unique_id
    #     ]
    # else:
    #     loss_node = []
    # if not targets:
    #     raise ValueError("no targets")
    # if len(node_info) < 1:
    #     raise ValueError("no node_info")
    # # if not loss_node:
    # #     raise ValueError("no loss_node")
    # return node_info, targets, loss_node
    
    # ### OLD RKGB END


def twremat_to_rockmate_schedule(K_graph, steps):
    kcn_id_to_node = {cnode.unique_id: cnode for cnode in K_graph.list_kcn}
    data_nodes = K_graph.list_kdn + [
        K_graph.input_kdn_grad,
        K_graph.input_kdn_data,
    ]

    print(f"OUTPUT SIZE TO KEEP ----{K_graph.output_kdn_data.mem}")

    op_list = []
    op_name_list = []
    alive_list = []
    alive_status = np.zeros(len(data_nodes), dtype=bool)
    alive_status[-1] = True  # input data tensor should be always alive!

    for i, step in enumerate(steps):
        step_type, cnode_id = step

        cnode_id = int(cnode_id)
        cnode = kcn_id_to_node[cnode_id]

        if step_type == "compute":
            ### Output data of the cnode computation should be stored
            for dnode in cnode.users:
                alive_status[data_nodes.index(dnode)] = 1
            alive_list.append(alive_status.copy())

            op = RunOp(cnode)
            op.no_grad = False
            op_list.append(op)
            op_name_list.append([op.name, op.op_type])

            # ### Input data to the cnode computation that won't be used by later computations should be deleted

            # # Get id of all computations that appear in the schedule after cnode
            # cnodes_compute_after = set([step[1] for step in list(filter(lambda step: step[0]=='compute', steps[i+1:]))])

            # # Get id of all data tensors that will serve as inputs to all computations performed after cnode
            # deps_cnodes_after = set()
            # for cn_id in cnodes_compute_after:
            #     cn = kcn_id_to_node[cn_id]
            #     for dn in set.union(cn.deps_real, cn.deps_fake):
            #     # for dn in set.union(cn.deps_real):
            #         deps_cnodes_after.add(dn.unique_id)
            # deps_cnodes_after.update([K_graph.input_kdn_data.unique_id])

            # # Free data tensors that served as inputs to cnode, but won't contribute to later computations
            # for dnode in data_nodes:
            #     idx = data_nodes.index(dnode)
            #     if alive_list[-1][idx] == 1 and not (dnode.unique_id in deps_cnodes_after):

            #         alive_status[idx] = 0
            #         alive_list.append(alive_status.copy())
            #         # print(alive_status)

            #         op = DelOp(data_nodes[idx])
            #         op.proxy = True
            #         op_list.append(op)
            #         op_name_list.append([op.name, op.op_type])

        elif step_type == "free":
            ### Twremat instruction "free computaion" we interpret as deleting data outputs of cnode computation
            for dnode in cnode.users:
                idx = data_nodes.index(dnode)
                if alive_status[idx]:
                    op = DelOp(data_nodes[idx])
                    op.proxy = True
                    op_list.append(op)
                    op_name_list.append([op.name, op.op_type])
                    alive_status[idx] = 0
                    alive_list.append(alive_status.copy())

        sched = OpSchedule(
            op_list,
            alive_list,
            K_graph.input_kdn_data,
            K_graph.input_kdn_grad,
            K_graph.output_kdn_data,
            K_graph.list_kdn,
            no_grad=False,
        )

    # from collections import Counter
    # counter = Counter(sched.op_name_list)
    # for k, v in counter.items():
    #     if 'fwd' in k:
    #         print(k, v, k.replace('fwd', 'bwd'), counter[k.replace('fwd', 'bwd')])

    # print(steps)
    # print(op_name_list)

    # print(f'Steps: {len(steps)},  op list: {len(sched.op_list)}')

    pred_mem = []
    for a, op in zip(sched.alive_list, sched.op_list):
        acc_mem = np.dot(a, sched.mem_sizes) - sched.input_size[1]

        pred_mem.append(acc_mem)
        if op.op_type == "Run":
            pred_mem[-1] += op.overhead

    print(pred_mem)
    print("peak_mem from op schedule " + str(np.max(np.array(pred_mem))))

    return sched



__all__ = ["solvers",
           "Rockmate",
           "PureRotor",
           "PureCheckmate",
           "PureRockmate",
           "Hiremate",
           "Offmate",
           "frontend",
           "generate_config",
           "from_config",
           "default_config",
           "save_config",
           "load_config"
           ]

from .rockmate import Rockmate
from . import solvers
from .frontend import *
from . import frontend



from distutils.core import setup, Extension
setup(
    name="rockmate_csolver",
    version="1.0",
    ext_modules=[
        Extension(
            "rockmate_csolver",
            ["solver.c"],
            extra_compile_args=["-O3", "--std=c99"],
        ),
    ],
)




__all__ = ["HILP", "RK_rotor", "TwRemat", "CheapSolver", "FastSolver"]

from .ilp.ilp_solver import HILP
from .cheap import CheapSolver
from .main import FastSolver
from .rk_rotor import RK_rotor

from .twremat import TwRemat



from rkgb.core.base import Node
from rkgb.core.backward import ComputationNode, AllocationNode
from rkgb.core.hierarchical import HierarchicalCluster
from collections import namedtuple
from copy import deepcopy
from typing import List
import warnings
import numpy as np
import torch
from rkgb.lowlevel.ast_add_on import make_str_list_assign


class Allocation:
    def __init__(
        self,
        target_name,
        alloc_type="",
        mem=0,
        info=dict(),
        dtype=torch.float32,
        size=None,
    ):
        """
        Allocation type should be in activation/parameters/buffer
        """
        self.target_name = target_name
        self._alloc_type = alloc_type
        self.mem = mem
        self.info = info
        self.dtype = dtype
        self.itemsize = dtype.itemsize if hasattr(dtype, "itemsize") else 4
        if size:
            self.size = round(size)
            self.mem = size * self.itemsize
        else:
            self.size = round(self.mem / self.itemsize)  # element size

    @property
    def name(self):
        return f"{self.target_name} {self.alloc_type}"

    @property
    def alloc_type(self):
        return self._alloc_type

    def __repr__(self):
        return self.name

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if isinstance(v, Node):
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))

        return result


class Activation(Allocation):
    def __init__(self, anode: AllocationNode, dtype=torch.float32):
        super().__init__(
            target_name=anode.main_target,
            alloc_type=anode.allocation_type,
            mem=anode.mem,
            info=anode.info,
            dtype=dtype,
        )
        self.kdn = anode
        self.anode = anode


class Parameter(Allocation):
    """
    There is not paramter grad/optim_states nodes from rkgb.
    Need to handle the dependency manually for Allocation.
    """

    def __init__(
        self, pnode, is_grad=False, is_optim_states=False, optimizer_states_factor=2
    ):
        self.pnode = pnode
        self.is_grad = is_grad
        self.is_optim_states = is_optim_states
        self.optimizer_states_factor = optimizer_states_factor
        if not self.is_optim_states:
            mem = pnode.mem
        else:
            mem = pnode.mem * self.optimizer_states_factor

        super().__init__(
            target_name=pnode.param_name,
            alloc_type="param",  # + "_grad"*grad+"_optim_states"*is_optim_states,
            mem=mem,
            info=pnode.info,
            dtype=pnode.info.dtype,
        )

    @property
    def param_name(self):
        return self.pnode.param_name

    @property
    def alloc_type(self):
        return "param" + "_grad" * self.is_grad + "_optim_states" * self.is_optim_states


class Buffer(Allocation):
    def __init__(self, name, mem=0, info=dict(), dtype=torch.float32, size=None):
        super().__init__(
            name=name, alloc_type="Buffer", mem=mem, info=info, dtype=dtype, size=size
        )
        self.dtype = dtype


class Op:
    def __init__(self, target_name, time=0, disabled=False, overhead=0):
        """
        Op type should be in Compute/Delete/Mapping/Allocate/Offload/Prefetch
        Compute/Delete/Mapping/Allocate happens in the main stream
        """
        self._target_name = target_name
        self.disabled = disabled
        self.overhead = overhead
        self._time = time
        self.fct_list = []
        self._op_type = type(self).__name__
        self.record_event = False
        self.wait_events = []

    def add_fct(self, f):
        self.fct_list.append(f)

    def __repr__(self):
        return f"{'Disabled_'*self.disabled}{self.name}"

    @property
    def target_name(self):
        return self._target_name

    @property
    def time(self):
        return self._time

    @property
    def name(self):
        return f"{self.op_type}({self.target_name})"

    @property
    def op_type(self):
        return self._op_type

    @op_type.setter
    def op_type(self, value):
        self._op_type = value

    def __call__(self):
        for f in self.fct_list:
            f()

    def __copy__(self):
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        return result

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for k, v in self.__dict__.items():
            if isinstance(v, Node):
                setattr(result, k, v)
            else:
                setattr(result, k, deepcopy(v, memo))

        return result

    def __eq__(self, op):
        if not type(self) == type(op):
            return False
        return op.name == self.name and self.disabled == op.disabled

    def __hash__(self):
        return id(self)

    def __getstate__(self):
        # fct_list should not be pickled
        state = self.__dict__.copy()
        state["fct_list"] = []
        return state


class SynchronizeOp(Op):
    def __init__(self, name="", disabled=False, stream=None):
        super().__init__(name, disabled=disabled)
        self.op_type = "Synchronize"
        self.stream = stream


class ComputeOp(Op):
    def __init__(
        self, cnode: ComputationNode, fast_forward=False, disabled=False, detach=True
    ):
        super().__init__(cnode.name, disabled=disabled)
        self.fast_forward = fast_forward
        self.detach = detach
        self.target: ComputationNode = cnode
        self.overhead = cnode.mem_overhead
        self.pos_info = {}  # positional information to be filled before compiling
        self.op_type = "Compute"
        self.time_mutiplier = 1# for dynamic batch

    @property
    def time(self):
        # cnode can be replaced during translation
        return self.target.time * self.time_mutiplier if self.target.time is not None else 0

    @property
    def target_name(self):
        return self.target.name


class DeleteOp(Op):
    def __init__(self, alloc: Allocation, disabled=False):
        super().__init__(alloc.name, disabled=disabled)
        self.target = alloc
        self.op_type = f"Delete"

    @property
    def target_name(self):
        return self.target.name


class MappingOp(Op):
    """
    The memory allocation of sources will be map to targets in buffer.
    the time of running this op is very short, but there is memory overhead.
    """

    def __init__(
        self,
        name: str,
        sources: list,
        targets: list,
        indices: list = None,
        disabled=False,
        copy=False,
    ):
        super().__init__("Mapping_" + name, disabled=disabled)
        self.sources = sources
        self.targets = targets
        self.indices = indices
        self.copy = copy
        self.overhead = sum(alloc.mem for alloc in sources)  # TODO: update when needed


class AllocateOp(Op):
    def __init__(self, alloc: Allocation, disabled=False, is_optim_states=False):
        super().__init__(alloc.name, disabled=disabled)
        self.target = alloc
        self.op_type = f"Allocate_{alloc.alloc_type}"
        self.is_optim_states = is_optim_states

    @property
    def target_name(self):
        return self.target.name


class OffloadOp(Op):
    def __init__(
        self,
        alloc: Allocation,
        indices: tuple = (0, None),
        disabled: bool = False,
        # grad: bool = False,
        # is_optim_states: bool = False,
        time: float = 0,
    ):
        super().__init__(alloc.name, disabled=disabled)
        self.target = alloc
        self.indices = indices
        self.disabled = disabled
        # self.grad = grad
        # self.is_optim_states = is_optim_states
        self._time = time
        self.op_type = f"Offload"

    @property
    def target_name(self):
        return self.target.name


class PrefetchOp(Op):
    def __init__(
        self,
        alloc: Allocation,
        indices: tuple = (0, None),
        disabled: bool = False,
        # is_optim_states: bool = False,
        time: float = 0,
    ):
        super().__init__(alloc.name, disabled=disabled)
        self.target = alloc
        self.indices = indices
        self.disabled = disabled
        # self.is_optim_states = is_optim_states
        self._time = time
        self.op_type = f"Prefetch"

    @property
    def target_name(self):
        return self.target.name


class OptimizeOp(Op):
    def __init__(
        self,
        list_params,
        cpu=False,
        alloc=None,
        disabled=False,
        time=0,
        overhead=0,
    ):
        self.list_params = list_params
        super().__init__(self.target_name, disabled=disabled, overhead=overhead)
        self.target = alloc or None
        self._time = time
        self.op_type = f"Optimize_{'cpu'*cpu}"
        self.is_cpu = cpu

    @property
    def target_name(self):
        return f"{','.join(self.list_params)}"


class ExecCodeOp(Op):
    def __init__(self, name, code, time=0, disabled=False, overhead=0):
        super().__init__(name, time, disabled, overhead)
        self.code = code


class PrepareOp(Op):
    # Prepare the placeholders for parameters before the iteration
    def __init__(
        self,
        alloc: Allocation,
        device: str = "cpu",
        cpu_placeholder=True,
        cpu_optimize=False,
        pin_memory=True,
        disabled: bool = False,
    ):
        super().__init__(alloc.name, disabled=disabled)
        self.target = alloc
        self.device = device
        self.cpu_placeholder = cpu_placeholder
        self.disabled = disabled
        self.op_type = f"Prepare_{alloc.alloc_type}"
        self.cpu_optimize = cpu_optimize
        self.pin_memory = pin_memory


class OpSchedule:
    solver = None

    def __init__(
        self,
        op_list: List[Op],
        cluster: HierarchicalCluster,
        loss_idx=None,
        with_parameters=False,
        init_alive_status: dict = {},
        init_op_list: list = [],
        restore_op_list: list = [],
        optimizer_states_factor=2,
        correct_overhead=True,
    ):
        """
        OpSchedule contains the operation list and automatically
        compute the alive status given the HierarchicalCluster.
        """
        self.op_list = op_list
        self.cluster = cluster
        self.loss_idx = loss_idx
        self.init_alive_status = init_alive_status
        self.init_op_list = init_op_list  # Place to prepare items in storage
        self.restore_op_list = restore_op_list
        self.with_parameters = with_parameters
        self.optimizer_states_factor = optimizer_states_factor
        self.interfaces = cluster.interfaces
        self.correct_overhead = correct_overhead
        self.with_optimization = any(isinstance(op, OptimizeOp) for op in op_list[::-1])

        self.create_list_alloc(cluster)
        self.get_sched_info()  # get the schedule information for higher level solving

    def __getstate__(self):
        # simulator should not be pickled
        state = self.__dict__.copy()
        state["simulator"] = None
        return state

    def get_occurrences(self):
        self.occurrences = dict()
        for i, op in enumerate(self.op_list):
            if op.name in self.occurrences:
                self.occurrences[op.name].append(i)
            else:
                self.occurrences[op.name] = [i]

    def is_occurred(self, op_name, i, next_i=None):
        if op_name not in self.occurrences:
            return False
        if next_i is None:
            return i in self.occurrences[op_name]
        else:
            return any(i <= i_ <= next_i for i_ in self.occurrences[op_name])

    def simulate_update(self, Simulator, refine_optimize=False):
        """
        Args:
            Simulator: Class defined to simulate op_sched
        """
        self.simulator = Simulator(self)
        self.simulator.refine()  # assume init_op_list remains the same
        if refine_optimize:
            self.simulator.refine_optimize()
        self.op_list = self.simulator.op_list
        self.steps = self.simulator.steps
        self.loss_step = self.simulator.loss_step
        self.loss_idx = self.simulator.loss_idx
        self.get_sched_info()

    def _sum_mem(self, alive_status_, ignore_list=[]):
        return sum(self.dict_alloc[k].mem
                   for k, v in alive_status_.items()
                   if k not in ignore_list and v)

    def get_memory(self, alive_list):
        L = len(self.op_list)
        self.time = np.zeros(L)
        self.save_mem = np.zeros(L)
        self.save_mem_with_interfaces = np.zeros(L)
        self.overhead = np.zeros(L)

        for i, (op, alive_status) in enumerate(zip(self.op_list, alive_list)):
            self.save_mem[i] = self._sum_mem(alive_status, ignore_list=self.interface_names)
            self.save_mem_with_interfaces[i] = self._sum_mem(alive_status)
            if op.disabled:
                continue
            self.time[i] = (
                op.time
                if not (isinstance(op, OffloadOp) or isinstance(op, PrefetchOp))
                else 0
            )
            self.overhead[i] = op.overhead

    def get_sched_info(self):
        """
        To get schedule information for higher level solving:
        - .mem: saved activation memory at the loss step, without interface activations
        - .fwd_time: time of forward operations
        - .bwd_time: time of backward operations
        - .fwd_overhead: during forward, the excess over the ending memory from the peak
        - .bwd_overhead: during backward, the excess over the ending memory from the peak
        - .phantoms: saved anodes at the loss step
        - .deps_interfaces_data: bwd requires interface data nodes which should be kept from fwd
        """
        self.get_occurrences()
        alive_list = self.create_alive_list()
        self.alive_list = alive_list
        self.get_memory(alive_list)

        self.mem = self.save_mem[self.loss_idx]
        self.peak_mem = max(self.save_mem + self.overhead)
        self.fwd_time = np.sum(self.time[: self.loss_idx + 1])
        self.bwd_time = np.sum(self.time[self.loss_idx + 1 :])

        def get_overhead_(save, overhead):
            return max(save + overhead) - save[-1]

        self.fwd_overhead = get_overhead_(
            self.save_mem[: self.loss_idx + 1],
            self.overhead[: self.loss_idx + 1],
        )
        if self.loss_idx < len(self.op_list) - 1:  # bwd is not empty
            self.bwd_overhead = get_overhead_(
                self.save_mem[self.loss_idx + 1 :],
                self.overhead[self.loss_idx + 1 :],
            ) - self.save_mem[self.loss_idx]
        else:
            self.bwd_overhead = 0

        self.phantoms = set()
        for anode in self.list_anodes:
            if (
                alive_list[self.loss_idx][anode.name]
                and not anode in self.all_interfaces
            ):
                self.phantoms.add(self.cluster.translate_representee_node(anode))

        self.dep_interfaces_data = set()
        for i, op in enumerate(self.op_list[self.loss_idx + 1 :]):
            if op.disabled or not isinstance(op, ComputeOp):
                continue
            for anode in op.target.deps_real:
                if anode in self.interfaces["input_data_anodes"]:
                    # self.dep_interfaces_data.add(self.list_anodes.index(anode))
                    self_anode = self.cluster.translate_representee_node(anode)
                    self.dep_interfaces_data.add(self_anode)
                if anode in self.interfaces["output_data_anodes"]:
                    for cnode in anode.deps:
                        if not self.is_occurred(
                            ComputeOp(cnode).name, self.loss_idx, self.loss_idx + i
                        ):
                            # if not generated during bwd
                            self_anode = self.cluster.translate_representee_node(anode)
                            # self.dep_interfaces_data.add(self.list_anodes.index(self_anode))
                            self.dep_interfaces_data.add(self_anode)

        self.fwd_overhead_correction = []
        self.bwd_overhead_correction = []
        if self.correct_overhead:
            self.add_correction_term(alive_list)

        self.offload_mem = sum(
            op.target.mem for op in self.op_list if isinstance(op, OffloadOp)
        )
        self.prefetch_mem = sum(
            op.target.mem for op in self.op_list if isinstance(op, PrefetchOp)
        )
        self.fwd_wait_time = 0
        self.bwd_wait_time = 0

    def add_correction_term(self, alive_list):
        """
        Correction term works as follows:
        each term corresponds to one step. "save" and "overhead" correspond
        to the inside anodes (non interfaces considered). 
        The other terms are the corrections for the higher level ILP:
        when the mark is "always", the interface will always appear in the overhead
        (still, they are recorded separately with the inside nodes);
        otherwise, the overhead will be written with the higher level ILP variables.
        """

        interfaces_status = []
        for anode in self.interfaces["input_data_anodes"]:  # Input of Fwd
            interfaces_status.append((anode.name, self.loss_idx))  # After fwd
            if anode in self.dep_interfaces_data:
                interfaces_status.append((anode.name, len(self.op_list)))  # After Bwd
        for anode in self.interfaces["output_data_anodes"]:  # Output of Fwd
            interfaces_status.append((anode.name, 0))  # Before fwd?
            if anode in self.dep_interfaces_data:
                interfaces_status.append((anode.name, len(self.op_list)))  # After Bwd
            else:
                interfaces_status.append((anode.name, -1))  # After Bwd

        for anode in self.interfaces["output_grad_anodes"]:
            interfaces_status.append((anode.name, len(self.op_list)))  # After Bwd
        for anode in self.interfaces["input_grad_anodes"]:
            interfaces_status.append((anode.name, self.loss_idx + 1))  # Before Bwd
        self.interfaces_status = interfaces_status
        for i, (op, alive_status) in enumerate(zip(self.op_list, alive_list)):
            if i == self.loss_idx:
                continue
            correction_term = {
                "save": self.save_mem[i], 
                "overhead": self.overhead[i],
            }
            for anode_name, index in interfaces_status:
                anode = self.dict_alloc[anode_name].anode
                if index == -1:
                    # special case: output_data in BWD without dependency
                    # If outside is alive, no need to correct;
                    # Otherwise, add anode to memory
                    if i > self.loss_idx and alive_status[anode_name] > 0:
                        correction_term[(self.list_anodes.index(anode), False)] = (
                            -anode.mem
                        )
                    continue

                if (
                    alive_status[anode_name] > 0
                    or (index > self.loss_idx) != (i > self.loss_idx)
                    # or not anode_name
                ):
                    # interfaces_status is useful when:
                    # 1. anode is not alive
                    # 2. Fwd to Fwd, Bwd to Bwd
                    continue

                if i >= index:  # if exist before
                    if (  # and not deleted in between
                        # anode_name not in self.op_name_list[index : i + 1]
                        self.is_occurred(DeleteOp(anode).name, index, i)
                    ):
                        correction_term[(self.list_anodes.index(anode), True)] = (
                            -anode.mem
                        )
                    else:
                        correction_term[(self.list_anodes.index(anode), "always")] = (
                            -anode.mem
                        )
                else:  # if exist afterwards
                    if not (anode in self.interfaces["output_data_anodes"]) and (
                        anode.deps
                        # and (list(anode.deps)[0].name in self.op_name_list[i : index + 1])
                        and self.is_occurred(
                            ComputeOp(list(anode.deps)[0]).name, i, index
                        )
                    ):  # and not generated in between
                        # check if output_data is created after i
                        correction_term[(self.list_anodes.index(anode), False)] = (
                            -anode.mem
                        )
                    elif anode in self.interfaces["input_data_anodes"]:
                        correction_term[(self.list_anodes.index(anode), False)] = (
                            -anode.mem
                        )
                    else:
                        correction_term[(self.list_anodes.index(anode), "always")] = (
                            -anode.mem
                        )

            if (
                i < self.loss_idx
                and correction_term not in self.fwd_overhead_correction
            ):
                self.fwd_overhead_correction.append(correction_term)
            elif (
                i > self.loss_idx
                and correction_term not in self.bwd_overhead_correction
            ):
                self.bwd_overhead_correction.append(correction_term)

    def create_list_alloc(self, cluster: HierarchicalCluster):
        self.all_interfaces = [
            anode for inter in self.interfaces.values() for anode in inter
        ]  # all interface KDN's
        self.interface_names = [anode.name for anode in self.all_interfaces]

        self.list_alloc = [Activation(anode) for anode in cluster.list_anodes]
        self.list_anodes = cluster.list_anodes
        if self.with_parameters:
            self.list_alloc.extend(
                [Parameter(anode) for anode in cluster.parameter_nodes]
            )
            self.list_alloc.extend(
                [
                    Parameter(anode, is_grad=True)
                    for anode in cluster.parameter_nodes
                    if anode.info.requires_grad
                ]
            )  # add parameter grad allocation
            if self.optimizer_states_factor:
                self.list_alloc.extend(
                    [
                        Parameter(
                            anode,
                            is_optim_states=True,
                            optimizer_states_factor=self.optimizer_states_factor,
                        )
                        for anode in cluster.parameter_nodes
                        if anode.info.requires_grad
                    ]
                )  # add parameter grad allocation
        self.dict_alloc_param = {
            alloc.name: alloc
            for alloc in self.list_alloc
            if isinstance(alloc, Parameter)
        }
        self.dict_alloc = {alloc.name: alloc for alloc in self.list_alloc}

    def create_alive_list(self):
        alive_status = {alloc.name: False for alloc in self.list_alloc}

        for alloc_name, is_alive in self.init_alive_status.items():
            alive_status[alloc_name] = is_alive

        for anode in self.interfaces["input_data_anodes"]:
            alive_status[anode.name] = True

        alive_list = []
        for i, op in enumerate(self.op_list):
            if op.disabled:
                if i == self.loss_idx:
                    for anode in self.interfaces["output_grad_anodes"]:
                        alive_status[anode.name] = True
                alive_list.append(alive_status.copy())
                continue
            if isinstance(op, DeleteOp):
                alive_status[op.target.name] = False
            elif isinstance(op, ComputeOp):
                # compute op should not be disabled except loss which is useful for alive status
                for anode in op.target.users:
                    if not ("phantoms" == anode.allocation_type and op.fast_forward):
                        alive_status[anode.name] = True
                if (
                    self.with_parameters and not op.target.is_fwd
                ):  # assume grad of parameters required by bwd will be generated
                    for pnode in op.target.required_parameter_nodes_real:
                        if pnode.info.requires_grad == True:
                            alive_status[Parameter(pnode, is_grad=True).name] = True
            elif isinstance(op, AllocateOp):
                alive_status[op.target.name] = True
            alive_list.append(alive_status.copy())
        # assert alive_status == self.init_alive_status# cyclic alive status
        # for k,v in self.init_alive_status.items():
        #     assert alive_status[k] == v
        return alive_list

    def add_pos_info(self):
        """
        Prepare positional information of each operation for compiling.
        """
        if not hasattr(self, "alive_list"):
            self.alive_list = self.create_alive_list()
        for index, op in enumerate(self.op_list):
            assert index in self.occurrences[op.name]
            if not isinstance(op, ComputeOp):
                continue
            op.pos_info = {
                "index": index,
                "first_occurrence": index == min(self.occurrences[op.name]),
                "last_occurrence": index == max(self.occurrences[op.name]),
            }
            if op.target.is_fwd:
                # last_before_bwd = True
                bwd_op_name = op.name.replace("FWD", "BWD")
                if bwd_op_name not in self.occurrences:
                    continue
                op.pos_info["next_bwd_idx"] = min(
                    [i for i in self.occurrences[bwd_op_name] if i > index],
                    default=None
                )
                if op.pos_info["next_bwd_idx"] is None:
                    op.pos_info["last_before_bwd"] = False
                else:
                    op.pos_info["last_before_bwd"] = not self.is_occurred(
                        op.name, index + 1, op.pos_info["next_bwd_idx"]
                    )

                # TODO: no_save_list should contain only the one got deleted before bwd
                cnode = op.target
                no_save_nodes = list(cnode.deps_real) + list(cnode.users)
                if self.with_parameters:
                    no_save_nodes += list(cnode.required_parameter_nodes_real)
                    no_save_nodes += list(cnode.required_parameter_nodes_fake)
                op.pos_info["no_save_list"] = [
                    (
                        anode.main_target
                        if hasattr(anode, "main_target")
                        else anode.param_name
                    )
                    for anode in no_save_nodes
                ]
            else:
                op.pos_info["temporary_tensor_names"] = []
                for anode in op.target.deps_fake:
                    if not self.alive_list[index][anode.name]:
                        op.pos_info["temporary_tensor_names"].append(anode.main_target)

                op.pos_info["input_names"] = []
                if not op.pos_info["first_occurrence"]:
                    # prev_i = index - self.op_list[:index][::-1].index(op) - 1
                    prev_i = max(i for i in self.occurrences[op.name] if i < index)
                    for anode in op.target.users:
                        if self.is_occurred(
                            DeleteOp(Activation(anode)).name, prev_i, index
                        ):
                            op.pos_info["input_names"].append(anode.main_target)

                    for pnode in op.target.required_parameter_nodes_real:
                        if self.is_occurred(
                            DeleteOp(Parameter(pnode)).name, prev_i, index
                        ):
                            op.pos_info["input_names"].append(pnode.param_name)
                    if not op.pos_info["input_names"]:
                        op.disabled = True
                        warnings.warn(f"{op.name} is recomputed but no target inputs")

    def __repr__(self) -> str:
        return f"Op_sched takes {sum(self.time):.2f} ms with {self.peak_mem/1024**2} MiB peak mem"



# ============
# = ROCKMATE =
# ============
import pickle
import torch
from torch import tensor
import time
from datetime import datetime
import warnings
import gc
import rkgb
from rkgb.lowlevel.preprocess_samples import ExampleInputs
from rkgb.lowlevel.measure import tensor_memory_size, get_Timer
from rkgb.lowlevel.constants import ref_verbose, ExceptionModuleDoesNotReqGrad
from rkgb.lowlevel.ast_add_on import ast_to_str, make_str_list_assign
from rkgb.core.partitioned import (
    PartitionerBottomToTop,
    PartitionerSequence,
    Partitioner,
)
from rkgb.lowlevel.constants import init_target_string

from .utils import get_optimize_metrics, measure_bandwidth
from .op_schedule import *
from .simulation import Simulator
from .solvers.main import preprocess, solve_recursive, FastSolver, add_sched
from .solvers import HILP, CheapSolver, RK_rotor
from .compiler import Compiler, RK_Storage, make_gd, Fct_record_event
import psutil



class Rockmate(torch.nn.Module):
    compiler = None
    autograd_Function = None
    backward_stop = False
    backward_add_output_grad = True
    op_sched = None
    module_does_not_req_grad = False

    def __init__(
        self,
        original_mod,
        model_inputs,
        budget=None,
        top_solvers=None,
        bottom_solvers=None,
        partitioners=None,
        rkgb_res=None,
        solve_sched=True,
        verbose=False,
        model_kwargs=None,
        cpu_optim=torch.optim.Adam,
        gpu_optim=torch.optim.Adam,
        optim_kwargs={},
        minor_offload_size=10 * 1024**2,
        keep_outputs=False,
        dynamic_batch_dim=None,
        solve_recursive=True,
    ):
        super().__init__()
        ref_verbose[0] = verbose
        self.top_solvers = top_solvers
        self.bottom_solvers = bottom_solvers
        if self.top_solvers is None and self.bottom_solvers is None:
            self.top_solvers = [HILP()]
            self.bottom_solvers = [HILP()]
        self.partitioners = partitioners
        if self.partitioners is None:
            self.partitioners = [PartitionerBottomToTop(can_use_rotor=False)]
        self.device = torch.device("cuda")  # Not obtaining from model
        self.exec_with_record_mem = False
        self.exec_with_record_time = False
        self.budget = budget
        self.rkgb_res = rkgb_res
        self.keep_outputs = keep_outputs
        self.list_solutions = []
        self.dynamic_batch_dim = dynamic_batch_dim
        self.timer = get_Timer(self.device)
        object.__setattr__(self, "original_mod", original_mod)

        self.config_partitioner()


        self.optimize_metrics = get_optimize_metrics(
            list(original_mod.parameters())[0],
            optim=gpu_optim,
            cpu_optim=cpu_optim,
            optim_kwargs=optim_kwargs,
            minor_offload_size=minor_offload_size,
        )

        self.get_rkgb_result(model_inputs, model_kwargs, minor_offload_size)

        # TODO: only measure bandwidth if needed
        self.bandwidth = measure_bandwidth()

        self.global_dict = make_gd(
            self.device,
            self.original_mod,
            self.dict_constants,
            optimize_metrics=self.optimize_metrics,
        )
        
        self.fix_solver_config()

        if solve_sched:
            self.solve_sched(recursive=solve_recursive)
            self.get_compiled_fct()

    def get_rkgb_result(self, model_inputs, model_kwargs, minor_offload_size):
        # -- use rkGB --
        try:
            if self.rkgb_res is None:
                self.rkgb_res = rkgb.rkgb.Result(
                    self.original_mod,
                    model_args=model_inputs,
                    model_kwargs=model_kwargs,
                    # verbose=verbose,
                    # wanted_graphs={"FB"},
                    partitioners=self.partitioners,
                    inspection_device=torch.device("cuda"),
                    print_time_in_each_stage=True,
                    dynamic_batch_dim=self.dynamic_batch_dim,
                )
            else:
                self.rkgb_res = self.rkgb_res

        except ExceptionModuleDoesNotReqGrad:
            self.module_does_not_req_grad = True

        self.dict_constants = self.rkgb_res.forward_and_backward_graph.dict_constants
        self.init_code = ast_to_str(self.rkgb_res.forward_and_backward_graph.init_code)
        self.output_names = [
            o.name
            for o in self.rkgb_res.forward_and_backward_graph.list_output_data_anodes
        ]

        self.minor_param_nodes = []
        for pnode in self.rkgb_res.hierarchical_cluster.parameter_nodes:
            if pnode.mem < minor_offload_size and not pnode.is_buffer:
                self.minor_param_nodes.append(pnode)
        for pnode in self.rkgb_res.S.init_node.required_parameter_nodes:
            if not pnode.is_buffer:
                self.minor_param_nodes.append(pnode)

    def config_partitioner(self):
        # ensure HILP config match partitioner config
        for partitioner in self.partitioners:
            if isinstance(partitioner, PartitionerBottomToTop):
                for solver in self.bottom_solvers:
                    if isinstance(solver, HILP):
                        if solver.config.nb_total_nodes < partitioner.config.max_estimate_per_sub_graph:
                            print(f"Warning, bottom solver HILP has nb_total_nodes {solver.config.nb_total_nodes}, "
                                  f"smaller than partitioner max_estimate_per_sub_graph {partitioner.config.max_estimate_per_sub_graph}."
                                  " This may result in failure to find schedules")
                for solver in self.top_solvers:
                    if isinstance(solver, HILP):
                        if solver.config.nb_total_nodes < partitioner.config.max_estimate_for_main_graph:
                            print(f"Warning, top solver HILP has nb_total_nodes {solver.config.nb_total_nodes}, "
                                  "smaller than partitioner max_estimate_for_main_graph "
                                  f"{partitioner.config.max_estimate_for_main_graph}. This may result in failure to find schedules")

    def fix_solver_config(self):
        # Set some options whoe values can only be known at runtime
        for solver in self.top_solvers:
            if isinstance(solver, HILP):
                solver.config.model_kwargs = solver.config.model_kwargs.copy()
                solver.config.protected_names = solver.config.protected_names.copy()
                solver.config.model_kwargs["optimize_metrics"] = self.global_dict["optimize_metrics"]
                solver.config.protected_names.extend([f"{init_target_string} data", f"{init_target_string} grad"])
                if self.keep_outputs:
                    solver.config.protected_names.extend(self.output_names)
                if self.dynamic_batch_dim is not None:
                    solver.config.model_kwargs["dynamic_batch_size"] = True

        for solver in self.bottom_solvers:
            if isinstance(solver, HILP):
                solver.config.protected_names.extend([f"{init_target_string} data", f"{init_target_string} grad"])
                if self.keep_outputs:
                    solver.config.protected_names.extend(self.output_names)


    def preprocess(self, solver = None):
        if solver is None:
            solver = FastSolver()
        for cluster in self.rkgb_res.hierarchical_structure.all_clusters:
            if not cluster.is_bottom:
                solver.preprocess(
                    cluster,
                    no_del_names=[
                        f"{init_target_string} data",
                        f"{init_target_string} grad",
                    ]
                    + self.output_names,
                )

    def fast_solve(self):
        solver = FastSolver()
        cluster = self.rkgb_res.hierarchical_cluster

        op_scheds = solver.solve(
                    cluster,
                    no_del_names=[
                        f"{init_target_string} data",
                        f"{init_target_string} grad",
                    ]
                    + self.output_names,
                )
        self.op_sched = op_scheds[0]

    def solver_recursive(self, list_solvers=None, only_preprocess=False):
        list_solvers = list_solvers or self.bottom_solvers

        solve_recursive(
            self.rkgb_res.hierarchical_cluster,
            list_solvers=list_solvers,
            skip_self=True,
        )

    def solve_sched(self, budget: float = None, list_solvers=None, recursive=True):
        # if multiple solvers in list_solvers,
        # will choose the one with minimum time
        budget = budget or self.budget
        # budget -= self.minor_size
        list_solvers = list_solvers or self.top_solvers

        # Set some options whoe values can only be known at runtime
        for solver in list_solvers:
            if isinstance(solver, HILP):
                # TODO: if no partitioning is allowed, update solver max nodes
                hilp_solver = True
                solver.config.optimize_metrics = self.global_dict["optimize_metrics"]
                if solver.config.offload:
                    if self.bandwidth is None:
                        self.bandwidth = measure_bandwidth()
                    solver.config.model_kwargs["bandwidth"] = self.bandwidth
        for solver in list_solvers:
            if isinstance(solver, HILP):
                solver.config.protected_names.extend([f"{init_target_string} data", f"{init_target_string} grad"])
                if self.keep_outputs:
                    solver.config.protected_names.extend(self.output_names)
                if self.dynamic_batch_dim is not None:
                    solver.config.model_kwargs["dynamic_batch_size"] = True

        self.preprocess()
        if self.bottom_solvers and recursive:
            self.solver_recursive()

        list_solutions = []
        for solver in list_solvers:
            list_solutions.extend(
                solver(self.rkgb_res.hierarchical_cluster, [budget])
            )

        self.rkgb_res.hierarchical_cluster.list_schedules.extend(list_solutions)
        if not list_solutions:
            warnings.warn("no feasible schedule is found")
            return False
        else:
            self.list_solutions.extend(list_solutions)
            self.op_sched = list_solutions[
                np.argmin([sum(op_sched.time) for op_sched in list_solutions])
            ]
            return True
            # self.op_list = self.op_sched.op_list

    def get_compiled_fct(self, new_compiler=True, simulate=True):
        if self.op_sched is None:
            return 
        if simulate:
            self.op_sched.simulate_update(Simulator, refine_optimize=True)
        if new_compiler:
            storage = RK_Storage()
            storage.init(self.global_dict)
            storage.dict_info = self.rkgb_res.forward_graph.dict_info
            self.compiler = Compiler(storage)
            if self.dynamic_batch_dim is not None:
                storage.dynamic_shapes = self.rkgb_res.raw_graph.dynamic_shapes

        self.minor_parameters = []
        if self.minor_param_nodes:
            self.minor_parameters = [
                pnode.get_value(self.original_mod)
                for pnode in self.minor_param_nodes
            ]

            def optimize():
                self.compiler.storage.ld["Optimize_minors"].step()
                for p in self.minor_parameters:
                    p.grad = None
                for pnode in self.minor_param_nodes:
                    code = make_str_list_assign(pnode.view_code, suffix=".data")
                    exec(code, self.global_dict, self.compiler.storage.ld)

            self.op_list[-1].add_fct(optimize)

        self.autograd_Function = define_autograd_Function(self)
        self.inherits_original_mod_attributes_and_methods()
        self.compiler.compile_preparation(
            self.rkgb_res.hierarchical_cluster,
            self.op_sched,
            self.minor_param_nodes,
            self.rkgb_res.forward_graph.output_nodes,
        )

        self.compiler.compile_sched(self.op_sched)

    def _exec(self, op: Op, disable=False):
        try:
            if self.exec_with_record_mem:
                torch.cuda.reset_peak_memory_stats()
                self.mem_before = torch.cuda.memory_allocated()
                self.max_before = torch.cuda.max_memory_allocated()
                if not disable:
                    op()
                torch.cuda.synchronize()
                allo_mem = torch.cuda.memory_allocated() - self.mem_before
                peak_mem = torch.cuda.max_memory_allocated() - self.max_before
                self.max_mem.append(peak_mem - allo_mem)
                self.allo_mem.append(allo_mem)
                op.allo_mem = allo_mem
                op.peak_mem = peak_mem - allo_mem
            else:
                if not disable:
                    op()
                else:
                    for f in op.fct_list:
                        if isinstance(f, Fct_record_event):
                            f()
        except Exception as e:
            print(f"Failed to execute {op}")
            raise e

    def exec_step(self, step):
        with torch.profiler.profile(
            activities=[
                torch.profiler.ProfilerActivity.CPU,
                torch.profiler.ProfilerActivity.CUDA,
            ],
            profile_memory=True,
        ) as p:
            self.timer.start()
            for op in step.op_list:
                self._exec(op)
            # torch.cuda.synchronize()
            # timer.end()
            # self.step_alloc_time.append(timer.elapsed())

            # timer.start()
            # for op in (
            #         step.ofl_ops
            #         +step.prf_ops
            #         +step.comp_ops
            #         +step.opt_ops
            #         +step.del_ops):
            #     self._exec(op)
            torch.cuda.synchronize()
            self.timer.end()
        p.export_chrome_trace(f"profiles/{len(self.step_profiles)}.json")
        self.step_time.append(self.timer.elapsed())
        self.step_memory_stats.append(torch.cuda.memory_stats())
        self.step_profiles.append(
            p.key_averages().table(sort_by="self_cuda_time_total", row_limit=-1)
        )

    def init_fwd_exec(self):
        # TODO: rename to first forward; also do the first backward differently incase grad not exist
        """
        The first forward execution to:
        1. allocate parameters in RK_storage;
        2. set placeholders for activations;
        3. run forward the first time so no parameter gradients.
        """

        for op in self.op_sched.init_op_list:
            self._exec(op)
        torch.cuda.synchronize()
        with torch.enable_grad():
            exec(self.init_code, self.global_dict, self.compiler.storage.ld)
            for op in self.op_list[: self.op_sched.loss_idx + 1]:
                disable = False
                if isinstance(op, OffloadOp) and isinstance(op.target, Parameter):
                    if op.target.is_grad or op.target.is_optim_states:
                        disable = True  # first iteration without grad offload
                if isinstance(op, OptimizeOp):
                    disable = True  # first iteration no need to optimize
                self._exec(op, disable=disable)
                # torch.cuda.synchronize()

    def restore_exec(self, keep_grad=False):
        """
        After the training iteration, restore parameters to original_mod.
        All the updated parameters of the original_mod will be kept on CPU.
        """

        if self.compiler.storage.ld == {}:
            return None
        self.zero_grad()
        for l in self.restore_fct_list:
            self._exec(l)
        for p in self.minor_parameters:
            p.data = p.data.to("cpu")

        for k, v in self.op_sched.dict_alloc_param.items():
            if v.pnode.is_buffer or v.pnode in self.minor_param_nodes:
                target = v.pnode.get_value(self.original_mod)
                target.data = target.data.to("cpu")
                continue

            if v.is_grad or v.is_optim_states:
                continue

            target = v.pnode.get_value(self.original_mod)
            target.data = self.compiler.storage.ld[v.target_name].data
            if keep_grad:
                target.grad = self.compiler.storage.ld[v.target_name].grad

        self.compiler.storage.ld = {}
        torch.cuda.synchronize()

    #  === nn.module's forward method wrapping self.autograd_Function.forward ===
    def forward(self, *args, record_mem=False, record_time=False, **kwargs):
        if self.module_does_not_req_grad:
            return self.original_mod(*args, **kwargs)
        self.exec_with_record_mem = record_mem
        self.exec_with_record_time = record_time
        self.max_mem = []
        self.allo_mem = []
        self.step_time = []
        self.step_alloc_time = []
        self.step_memory_stats = []
        self.step_profiles = []
        if not self.training:
            self.original_mod.eval()
            return self.original_mod(*args, **kwargs)
        else:
            if self.compiler is None:
                raise Exception(
                    "No schedule compiled, no solution to exec,\n"
                    "To be able to use the RkMod you first need "
                    "to call get_compiled_fct(_HILP)."
                )
            elif self.autograd_Function is None:
                raise Exception(
                    "The custom forward and backward functions haven't "
                    "been generated yet, please call the function: "
                    "define_autograd_Function"
                )
            # -> Send the inputs to Function.forward via the buffer (Rem 1)
            self.inputs_buffer = ExampleInputs(self.original_mod, args, kwargs)
            self.dict_inputs_buffer = dict_inputs = self.inputs_buffer.dict

            # -> Pass the inputs which req grad to prepare their backward (Rem 1)
            inputs_which_req_grad = []
            name_of_inputs_which_req_grad = []
            for k, v in dict_inputs.items():
                if isinstance(v, torch.Tensor) and v.requires_grad:
                    inputs_which_req_grad.append(v)
                    name_of_inputs_which_req_grad.append(k)
            self.name_of_inputs_which_req_grad_buffer = name_of_inputs_which_req_grad

            dummy_input = torch.ones(1).requires_grad_()
            output_mt_values = self.autograd_Function.apply(
                dummy_input, *inputs_which_req_grad
            )

            self.out_code = []
            self.output_targets = []
            # for output_node in self.rkgb_res.simplified_graph.output_nodes:
            # for main_target, set_output_targets in self.rkgb_res.simplified_graph.dict_output_mt_to_targets_sent.items():
            for main_target in self.rkgb_res.forward_and_backward_graph.output_targets:
                # anode = self.rkgb_res.hierarchical_cluster.dict_nodes[main_target]
                cnode = self.rkgb_res.hierarchical_cluster.dict_nodes[
                    f"FWD[{main_target}]"
                ]
                code = make_str_list_assign(cnode.body_code)
                for out_target in cnode.all_targets:
                    code = code.replace(out_target, f"out_{out_target}")
                    self.output_targets.append(f"out_{out_target}")

                # for output_target in set_output_targets:
                #     # ast_code = self.rkgb_res.simplified_graph.dict_output_viewing_code[main_target]
                #     # code = ""
                #     # for assign in ast_code.body:
                #     #     if ast_to_str(assign.targets) == output_target:
                #     #         code += f"{ast_to_str(assign)}\n"

                #     code = code.replace(output_target, f"out_{output_target}")
                #     if main_target != output_target:
                #         code = code.replace(main_target, f"out_{main_target}")
                #     else:
                #         code = ""
                self.out_code.append(code)
                exec(code, self.compiler.storage.gd, self.compiler.storage.ld)

            # -> Output what's defined in module
            output_nodes = self.rkgb_res.forward_graph.output_nodes
            final_outputs = [
                self.compiler.get_val(f"out_{output.main_target}")
                for output in output_nodes
            ]
            return tuple(final_outputs)

    def zero_grad(self, set_to_none=True, remain_for_offload=True):
        if remain_for_offload:
            remains = []
            # for k,v in self.op_sched.dict_alloc_param.items():
            #     if v.is_grad and self.op_sched.alive_list[-1][k]:
            #     # if v.is_grad and self.op_sched.init_alive_status[k]:
            #         remains.append(v.pnode.param_name)
            for k, p in self.original_mod.named_parameters():
                if k not in remains and p.grad is not None:
                    # if set_to_none:
                    #     p.grad = None
                    # else:
                    p.grad.zero_()
        else:
            self.original_mod.zero_grad(set_to_none=set_to_none)

    @property
    def minor_size(self):
        optimizer_states_factor = 0
        if self.optimize_metrics:
            optimizer_states_factor += self.optimize_metrics["optimizer_states_factor"]
        return sum([pnode.mem for pnode in self.minor_param_nodes]) * (optimizer_states_factor+1)

    @property
    def op_list(self):
        return self.op_sched.op_list

    def inherits_original_mod_attributes_and_methods(self):
        for k, v in self.original_mod.__dict__.items():
            if not "forward" in k and not "backward" in k and k not in ["training"]:
                self.__dict__[k] = v

    def save_to_local(self, path, id=None):
        if not id:
            id = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")    ## rkMod.save_sched_to_local("dump")
        with open(f"{path}/{id}_rkgb_res.pkl", "wb") as f:
            pickle.dump(self.rkgb_res, f)

    def save_sched_to_local(self, path, id=None):
        if not id:
            id = datetime.now().strftime("%d_%m_%Y_%H_%M_%S")
        with open(f"{path}/{id}_sched.pkl", "wb") as f:
            pickle.dump(self.op_sched, f)

    def load_from_local(self, path, id, load_sched=True):
        with open(f"{path}/{id}_rkgb_res.pkl", "rb") as f:
            self.rkgb_res = pickle.load(f)
            if load_sched:
                self.list_solutions = self.rkgb_res.hierarchical_cluster.list_schedules
                try:
                    with open(f"{path}/{id}_sched.pkl", "rb") as f_sched:
                        self.op_sched = pickle.load(f_sched)
                except Exception:
                    self.op_sched = self.list_solutions[
                        np.argmin(
                            [sum(op_sched.time) for op_sched in self.list_solutions]
                        )
                    ]
                self.get_compiled_fct()

    def load_sched_from_local(self, path, id):
        with open(f"{path}/{id}_sched.pkl", "rb") as f_sched:
            self.op_sched = pickle.load(f_sched)
        self.get_compiled_fct()


def define_autograd_Function(RkMod: Rockmate):
    class RK_autograd_Function(torch.autograd.Function):
        # === OUR FORWARD FUNCTION ===
        @staticmethod
        def forward(ctx, dummy_input, *args):
            if RkMod.compiler.storage.ld != {}:
                ctx.RK_Storage = storage = RkMod.compiler.storage
                ctx.name_of_inputs_which_req_grad = (
                    RkMod.name_of_inputs_which_req_grad_buffer
                )
                dict_inputs = (
                    RkMod.dict_inputs_buffer
                )  # args are passed in RkMod.forward
                #  -> Detach input tensors (Rem 3) and store all the inputs
                dict_input_tensors_detach = dict()  #  dict : input -> detached input
                for k, v in dict_inputs.items():
                    if isinstance(v, torch.Tensor):
                        v_d = v.detach().requires_grad_(v.requires_grad)
                        dict_input_tensors_detach[v] = v_d
                        storage.ld[k] = v_d
                    #  TODO elif iterables of Tensors ?
                    else:
                        storage.ld[k] = v

                with torch.enable_grad():
                    exec(
                        RkMod.init_code, RkMod.global_dict, storage.ld
                    )  # is compiler.global_dict
                    if RkMod.exec_with_record_time and RkMod.op_sched.loss_step > 0:
                        for step in RkMod.op_sched.steps[: RkMod.op_sched.loss_step]:
                            RkMod.exec_step(step)
                    else:
                        for op in RkMod.op_list[: RkMod.op_sched.loss_idx]:
                            RkMod._exec(op)
            else:
                # *** INITIALIZATION PART ***
                #  -> Get the inputs using the buffer (Rem 1)
                dict_inputs = RkMod.dict_inputs_buffer
                RkMod.dict_inputs_buffer = None

                #  -> Create the RK_Storage for this run, and store it in ctx
                ctx.RK_Storage = storage = RkMod.compiler.storage
                storage.init(RkMod.global_dict)
                RkMod.compiler.storage = storage

                #  -> Store what we need to return inputs' grad (Rem 1)
                ctx.name_of_inputs_which_req_grad = (
                    RkMod.name_of_inputs_which_req_grad_buffer
                )

                #  -> Detach input tensors (Rem 3) and store all the inputs
                dict_input_tensors_detach = dict()  #  dict : input -> detached input
                for k, v in dict_inputs.items():
                    if isinstance(v, torch.Tensor):
                        v_d = v.detach().requires_grad_(v.requires_grad)
                        dict_input_tensors_detach[v] = v_d
                        storage.ld[k] = v_d
                        if RkMod.dynamic_batch_dim is not None:
                            inspect_batch = storage.dict_info[k].tensor_size[
                                RkMod.dynamic_batch_dim
                            ]
                            batch_multiplier = (
                                v_d.shape[RkMod.dynamic_batch_dim] / inspect_batch
                            )
                            storage.batch_multiplier = batch_multiplier

                    #  TODO elif iterables of Tensors ?
                    else:
                        storage.ld[k] = v

                torch.cuda.synchronize()
                with torch.enable_grad():
                    RkMod.init_fwd_exec()
                torch.cuda.synchronize()

            #  *** EXECUTION PART ***
            # -> Autograd turns off itself before giving use the control.
            # -> But we need it to forward/backward each node.

            # -> Get the output
            outs = []
            for (
                out_node
            ) in RkMod.rkgb_res.forward_and_backward_graph.list_output_data_anodes:
                out_mt = out_node.main_target
                # if out_mt == RkMod.rkgb_res.forward_graph.nodes[-1].main_target:
                #     # Last output node; its grad_fn will not be called again
                #     outs.append(RkMod.compiler.get_val(out_mt))
                #     continue
                RkMod.compiler.get_val(f"out_{out_mt}").data = RkMod.compiler.get_val(
                    out_mt
                )
                o = RkMod.compiler.get_val(f"out_{out_mt}")
                outs.append(o)

            if len(outs) == 1:
                return outs[0]
            else:
                return tuple(outs)
            # -> Remember that out have been detached from the rest during exec
            """
            ctx.set_materialize_grads(True) # as the default
            # -> so we don't have to check if grad_output is None
            # -> if outputs' grad is None Autograd fill them with zeros
            """

        # === END OF FORWARD FUNCTION ===

        """
        @staticmethod
        def setup_context(ctx,inputs,outputs):
            pass
        # PyTorch prefer to handle the ctx in a separate method,
        # but it makes more sense for us to handle ctx during the forward.
        # (Also, setup_context can only access to inputs and outputs,
        # so they suggest to pass intermediate values as outputs,
        # but outputs can only be tensors)
        # Anyway, it's not important, Autograd is originally designed
        # to handle the ctx during the forward, it's just that PyTorch 2.0
        # now prefers to use this separate method.
        """

        # === OUR BACKWARD FUNCTION ===
        @staticmethod
        @torch.autograd.function.once_differentiable
        def backward(ctx, *grad_outs):
            #  -> Reload the storage and out
            # storage = ctx.RK_Storage
            # RkMod.compiler.storage = storage
            # -> Put grad_out in out.grad (Rem 4)
            for out_node, out_grad in zip(
                RkMod.rkgb_res.forward_and_backward_graph.list_output_data_anodes,
                grad_outs,
            ):
                out = RkMod.compiler.get_val(out_node.main_target)
                # if out_grad.mean() != 0:
                out.grad = out_grad.data.as_strided_(
                    out.shape, out.stride(), out.storage_offset()
                )
                # print(out_node.main_target, out.grad.mean)
                out_grad.data = torch.empty(0)

            if not RkMod.keep_outputs:
                for target in RkMod.output_targets:
                    RkMod.compiler.storage.ld[target].data = torch.empty(0)
                    RkMod.compiler.storage.ld[target].grad = None

            loss_idx = RkMod.op_sched.loss_idx
            #  * record_mem stuff *
            # if RkMod.exec_with_record_mem:
            #     RkMod.output_size = tensor_memory_size(
            #         storage.ld[RkMod.output.main_target]
            #     )
            #     # TODO: record output grad if needed

            with torch.enable_grad():
                stop = RkMod.backward_stop
                if stop:
                    for i, op in enumerate(
                        RkMod.op_list[loss_idx + 1 : loss_idx + stop + 1]
                    ):
                            RkMod._exec(op)
                else:
                    if RkMod.exec_with_record_time:
                        for step in RkMod.op_sched.steps[RkMod.op_sched.loss_step :]:
                            RkMod.exec_step(step)
                    else:
                        for op in RkMod.op_list[loss_idx + 1 :]:
                            RkMod._exec(op)
                    # if RkMod.exec_with_record_mem and RkMod.backward_add_output_grad:
                    #     RkMod.allo_mem[loss_idx] += RkMod.output_size
                    #  -> return grad of dummy input + inputs' which req grad (Rem 1)
                    grad_inputs = tuple(
                        RkMod.compiler.get_val(inp).grad
                        for inp in ctx.name_of_inputs_which_req_grad
                    )
                    grads = (torch.ones(1),) + grad_inputs

                    return grads

        # === END OF BACKWARD FUNCTION ===

    return RK_autograd_Function



from distutils.core import setup, Extension
setup(
    ext_modules=[
        Extension(
            "rockmate.solvers.rk_rotor.csolver",
            ["src/rockmate/solvers/rk_rotor/solver.c"],
            extra_compile_args=["-O3", "--std=c99"],
        ),
    ],
)




# from rkgb.Htools import H_cluster
from rkgb.core.hierarchical import HierarchicalCluster
from .twremat_utils import runtwremat, get_twremat_graph
from ..op_schedule import OpSchedule, ComputeOp, DeleteOp, Activation
from .main import Solver, get_cluster_budget
from dataclasses import dataclass


class TwRemat(Solver):
    '''This solver is a wrapper for the TwRemat treewidth-based rematerialization algorithm.

    The TwRemat algorithm is implemented in https://github.com/lemonlabsuk/ai-scala-developer
    and described in the paper

    Ravi Kumar, Manish Purohit, Zoya Svitkina, Erik Vee, Joshua Wang. Efficient Rematerialization
    for Deep Networks, NeurIPS
    2019. https://papers.nips.cc/paper/9653-efficient-rematerialization-for-deep-networks.pdf

    This class uses an external twremat executable that should be available in the PATH. A
    singularity image containing all requirements is linked in the project README.
    '''

    @dataclass
    class Config:
        mem_unit: int = 1024**2
        contains_data_node: bool = False
        allow_loss_recomputation: bool = False
        verbose: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _can_solve(self, cluster: HierarchicalCluster):
        return not cluster.is_bottom

    def solve(self, cluster, budgets=None):
        if not self._can_solve(cluster):
            return []
        if budgets is None:
            budgets = get_cluster_budget(
                cluster.representee_cluster, with_save_budget=False
            )
        else:
            budgets = budgets
        list_sched = []
        for budget in budgets:
            if isinstance(cluster, HierarchicalCluster):
                list_sched.append(self.solve_cluster(cluster, int(budget))[1])

        return list_sched

    def solve_cluster(self, cluster, budget):
        node_info, target, loss = get_twremat_graph(
            cluster,
            self.config.contains_data_node,
            self.config.allow_loss_recomputation,
        )
        steps = runtwremat(node_info, budget, target, loss, verbose=self.config.verbose)

        kcn_id_to_node = {cnode.unique_id: cnode for cnode in cluster.list_cnodes}
        data_nodes = cluster.list_anodes

        op_list = []
        op_name_list = []
        loss_idx = None

        for i, step in enumerate(steps):
            step_type, cnode_id = step

            cnode_id = int(cnode_id)
            cnode = kcn_id_to_node[cnode_id]

            if step_type == "compute":
                if "loss" in cnode.name and loss_idx is None: #for multiple loss nodes, we need the first one
                    loss_idx = len(op_list)

                runOp = ComputeOp(cnode, disabled=("loss" in cnode.name))
                op_list.append(runOp)
                op_name_list.append(runOp.name)

                # ### Input data to the cnode computation that won't be used by later computations should be deleted

                # # Get id of all computations that appear in the schedule after cnode
                # cnodes_compute_after = set([step[1] for step in list(filter(lambda step: step[0]=='compute', steps[i+1:]))])

                # # Get id of all data tensors that will serve as inputs to all computations performed after cnode
                # deps_cnodes_after = set()
                # for cn_id in cnodes_compute_after:
                #     cn = kcn_id_to_node[cn_id]
                #     # for dn in set.union(cn.deps_real, cn.deps_fake):
                #     for dn in set.union(cn.deps_real):
                #         deps_cnodes_after.add(dn.unique_id)
                # deps_cnodes_after.update([cluster.list_kdn.unique_id])

                # # Free data tensors that served as inputs to cnode, but won't contribute to later computations
                # for dnode in data_nodes:
                #     idx = data_nodes.index(dnode)
                #     if not (dnode.unique_id in deps_cnodes_after):
                #         op_list.append(Op(data_nodes[idx]))
                #         op_name_list.append(op.name)

            elif step_type == "free":
                ### Twremat instruction "free computaion" we interpret as deleting data outputs of cnode computation
                for dnode in cnode.users:
                    idx = data_nodes.index(dnode)
                    delOp = DeleteOp(Activation(data_nodes[idx]))
                    op_list.append(delOp)
                    op_name_list.append(delOp.name)

        sched = OpSchedule(op_list, cluster=cluster, loss_idx=loss_idx)
        sched.solver = "TWRemat"
        return op_name_list, sched



from typing import Dict, Any
import time
import numpy as np
from copy import deepcopy
from ...op_schedule import (
    Activation,
    Parameter,
    Buffer,
    ComputeOp,
    DeleteOp,
    MappingOp,
    AllocateOp,
    OffloadOp,
    PrefetchOp,
    SynchronizeOp,
    OptimizeOp,
    PrepareOp,
    ExecCodeOp,
    OpSchedule,
)
from ..main import get_sched, add_sched, translate
from .ilp_model import ModelPULP
from .ilp_offload import ModelPULPOffload


class knapsack:
    def __init__(self, parameter_sizes: list, pre_solve_size=10):
        size = [s[1] for s in parameter_sizes]
        self.parameter_sizes = parameter_sizes
        self.sizes = [s / sum(size) for s in size]
        self.pre_solve_size = pre_solve_size

    def get_size(self, indices, sizes):
        return sum(sizes[i] for i in indices)

    # @lru_cache(maxsize=4096 * 4096)
    def solve(self, frac: float, i: int = 0, sizes=[]):
        sizes = sizes or self.sizes
        if frac < 0:
            return []
        if i == len(sizes):
            return list(range(i))
        res1 = self.solve(frac, i + 1, sizes)
        res2 = self.solve(frac - sizes[i], i + 1, sizes)
        res2 = [i] + res2
        if self.get_size(res1, sizes) <= self.get_size(res2, sizes):
            return res1
        else:
            return res2

    def select(self, frac: float):
        sizes = self.sizes.copy()
        parameter_sizes = self.parameter_sizes.copy()
        selections = []
        while len(sizes) > self.pre_solve_size and frac > 0:
            sel_i = self.presolve(frac, sizes)
            if sel_i is None:
                break
            selections.append(parameter_sizes.pop(sel_i)[0])
            frac -= sizes.pop(sel_i)
        indices = self.solve(frac, sizes=sizes)
        selections += [parameter_sizes[i][0] for i in indices]
        return selections

    def select_size(self, size: int):
        if not self.parameter_sizes:
            return []
        return self.select(size / sum(s[1] for s in self.parameter_sizes))

    def presolve(self, frac, sizes):
        array = np.array(sizes)
        sel_i = np.argmax(array * (array < frac))
        if array[sel_i] > frac:
            return np.argmin(array)
        return sel_i


def schedule(md: ModelPULP, hgraph=None, check_valid=False):
    """
    Given the solution from HILP, we want to translate the result
    to a OpSchedule that can be used in a higher level.
    """
    hgraph = hgraph if hgraph else md.hgraph

    init_op_list = []
    restore_op_list = []
    init_alive_status = {}
    loss_op = ComputeOp(md.hgraph.cluster.loss_cnode, disabled=True)
    if isinstance(md, ModelPULPOffload):
        W = len(md.parameter_size)
        (
            op_list,
            init_alive_status,
            init_op_list,
            restore_op_list,
        ) = schedule_offload(md, hgraph)
        op_name_list = [op.name for op in op_list]

        init_ops = {op.target.name: op for op in init_op_list}
        for pnode in md.hgraph.cluster.parameter_nodes:
            if pnode.mem < md.minor_offload_size or pnode.is_buffer:
                device = "cuda"
            else:
                device = "cpu"
            alloc = Parameter(pnode)
            alloc_grad = Parameter(pnode, is_grad=True)
            if alloc.name not in init_ops:
                init_op_list.append(
                    PrepareOp(
                        alloc,
                        device=device,
                        cpu_optimize=OffloadOp(alloc_grad).name
                        in op_name_list,
                        pin_memory=OffloadOp(alloc).name in op_name_list,
                    )
                )

    else:
        op_list = []
        for pnode in md.hgraph.cluster.parameter_nodes:
            init_op_list.append(ExecCodeOp(pnode.param_name, pnode.get_code()))
        for t in range(md.T):
            for k in md.krange(t):
                if t == md.loss_idx and k == md.loss_idx:
                    op_list.append(loss_op)
                op_list += schedule_compute(md, t, k, hgraph)

    # print("finish scheduling")
    for anode in md.hgraph.cluster.interfaces["input_data_anodes"]:
        init_alive_status[anode.name] = True  # anode share the name as alloc

    op_sched = OpSchedule(
        op_list,
        loss_idx=op_list.index(loss_op),
        cluster=md.hgraph.cluster,
        init_alive_status=init_alive_status,
        init_op_list=init_op_list,
        restore_op_list=restore_op_list,
        with_parameters=isinstance(md, ModelPULPOffload),
        optimizer_states_factor=(
            md.optimize_metrics["optimizer_states_factor"]
            if hasattr(md, "optimize_metrics")
            else None
        ),
    )
    # check_valid = True
    if check_valid:
        for op, alive_status in zip(op_sched.op_list, op_sched.alive_list):
            if op.is_del:
                continue
            for kdn in op.kn.deps_real:
                if not alive_status[kdn.name]:
                    print(f"Invalid sched found: try to run {op.kn} without {kdn}")
                    raise ValueError
    return op_sched


def schedule_compute(md: ModelPULP, t, k, hgraph):
    op_list = []
    sol = md.sol

    j = md.hcn2sub_c[k]
    # if md.sumComp[t, k].value() == 1:
    if sol(md.sumComp[t, k]):
        hcn = hgraph.list_HCNs[k]
        opt = -1
        for o in range(md.nSched[k]):
            if sol(md.Comp[t, k, o]):
                opt = o
                break
        if opt > -1:
            h_obj = md.list_list_sched[j][opt]
            if hcn.is_fwd:
                sub_op_list = h_obj.op_list[: h_obj.loss_idx]
            else:
                sub_op_list = h_obj.op_list[h_obj.loss_idx + 1 :]

                # if md.sumAliveP[(j, t + 1)].value() == 0:
                # sub_op_list.append()
            sub_op_list = deepcopy(sub_op_list)

            if (
                not hcn.is_fwd
                # and md.sumAliveP[(j, t + 1)].value() > 0
                and sol(md.sumAliveP[t + 1, j])
            ):  # phantoms should be kept
                phantoms_to_keep = h_obj.phantoms
                # for op in sub_op_list[::-1]:
                #     if (
                #         op.is_del
                #         and not op.disabled
                #         and op.kn in phantoms_to_keep
                #     ):
                #         # Only the last del should be disabled
                #         op.disabled = True
                #         phantoms_to_keep.remove(op.kn)
            # translating sub_op_list
            if hcn.sub_cluster is not hcn.sub_cluster.representee_cluster:
                sub_op_list = translate(hcn.sub_cluster, sub_op_list)
        else:
            h_obj = hcn
            sub_op_list = deepcopy(h_obj.ff_op_list)

        anode_to_protect = []
        if not hcn.sub_cluster is None:
            for anode in hcn.sub_cluster.all_interfaces:
                i_ = [hdn.anode.name for hdn in md.hgraph.list_HANs].index(anode.name)
                if (k, i_) in md.delete_list and k in md.han_users_real[i_]:
                    eidx = md.delete_list.index((k, i_))
                    if not sol(md.delete[t, eidx]):
                        anode_to_protect.append(anode.name)

        for op in sub_op_list:
            if isinstance(op, DeleteOp) and op.target.anode.name in anode_to_protect:
                op.disabled = True
        op_list += sub_op_list

    for eidx, (k_, i) in enumerate(md.delete_list):
        # print(k_, i)
        # if k == k_ and md.delete[t, eidx].value()==1:
        if k == k_ and sol(md.delete[t, eidx]):
            han = hgraph.list_HANs[i]
            op_list.append(DeleteOp(Activation(han.anode)))
    return op_list


def schedule_offload(md: ModelPULPOffload, hgraph=None):
    """
    V1: md.grouping = False:
    merge every cluster, ofl/prf/del partially, high memory overhead
    """
    hgraph = hgraph if hgraph else md.hgraph

    ### Handle multiplier
    md.params_vars = [
        md.AliveW,
        md.OflWProg,
        md.OflW,
        md.PrfW,
        md.PrfWProg,
        md.OptC,
        md.AliveG,
        md.OflGProg,
        md.OflG,
        md.PrfG,
        md.PrfGProg,
        md.AliveO,
        md.OflO,
        md.OflOProg,
        md.PrfO,
        md.PrfOProg,
    ]
    if isinstance(md.req_w, float):
        multiplier = 1 - md.req_w
    else:
        multiplier = 1 - md.req_w.value()
    for p in md.params_vars:
        for k, v in p.items():
            p[k] = v * 1 / (1 - multiplier)
            pass

    if md.activation_offload:
        md.selected_phantoms = {}

    def add_op(op_dict, op_list):
        for op in op_list:
            op_dict[op[:2]].append(op[2])

    md.ofl_ops = {step: [] for step in md.active_steps}
    md.prf_ops = {step: [] for step in md.active_steps}
    md.del_ops = {step: [] for step in md.active_steps}
    md.opt_ops = {step: [] for step in md.active_steps}
    md.cpu_optimized_params = {}
    md.cpu_optimized_steps = {step: [] for step in md.active_steps}
    init_op_list = []
    restore_op_list = []
    init_alive_status = dict()
    if md.grouping:
        for w in range(md.W)[::-1]:
            o_l, p_l, d_l, t_l, i_l, r_l, init_alive = group(md, w)
            add_op(md.ofl_ops, o_l)
            add_op(md.prf_ops, p_l)
            add_op(md.del_ops, d_l)
            add_op(md.opt_ops, t_l)
            init_op_list.extend([ops[2] for ops in i_l])
            restore_op_list.extend([ops[2] for ops in r_l])
            for alloc in init_alive:
                init_alive_status[alloc.name] = True

        for j in range(md.J):
            ofl_ops, prf_ops, del_ops = group_activation_offload(md, j)
            add_op(md.ofl_ops, ofl_ops)
            add_op(md.prf_ops, prf_ops)
            add_op(md.del_ops, del_ops)
    # else:
    #     init_op_list = md.schedule_init_op_list()

    sol = md.sol
    # offload_buffers = {w:[] for w in range(W)}
    op_list = []

    # for op in init_op_list:
    #     if isinstance(op, AllocateOp):
    #         init_alive_status[op.target] = True

    for t in range(md.T):
        for k in md.krange(t):
            op_list.extend(schedule_step(md, t, k))

    return op_list, init_alive_status, init_op_list, restore_op_list

def schedule_step(md: ModelPULP, t, k):
    op_list = []
    if not md.sol(md.sumComp[t, k]):
        return op_list
    op_list.append(SynchronizeOp(f"{(t,k)}"))
    if t == md.loss_idx and k == md.loss_idx:
        op_list.append(ComputeOp(md.hgraph.cluster.loss_cnode, disabled=True))
    
    j = md.hcn2sub_c[k]
    op_list += schedule_compute(md, t, k, md.hgraph)
    last_op = op_list[-1]
    last_op.record_event = True
    self_ops = []
    self_targets = [
        p.param_name for w in md.hcn2param[k] for p in md.parameters[w]
    ]
    if md.hcn2sub_c[k] is not None and md.activation_offload:
        self_targets += md.selected_phantoms[j]
    for op in md.opt_ops[t, k]:
        op_list.append(op)
        if op.target.target_name in self_targets:
            op.record_event = True
    for op in md.ofl_ops[t, k]:
        if op.target.target_name in self_targets:
            self_ops.append(op)
        else:
            op_list.append(op)
    for op in md.del_ops[t, k]:
        if op.target.target_name in self_targets:
            self_ops.append(op)
        else:
            op_list.append(op)
    for op in md.prf_ops[t, k]:
        op_list.append(op)

    op_list.extend(self_ops)

    occurrences = dict()
    for i, op in enumerate(op_list):
        if op.name not in occurrences:
            occurrences[op.name] = []
        occurrences[op.name].append(i)
    for op in self_ops:
        if isinstance(op, OffloadOp):
            op.record_event = True
            opt_op_name = OptimizeOp([op.target.pnode.param_name],
                        alloc=op.target).name
            if opt_op_name in occurrences:
                opt_op = op_list[max(occurrences[opt_op_name])]
                op.wait_events.append((opt_op.op_type, opt_op.target.name))

        if isinstance(op, DeleteOp):
            ofl_op_name = OffloadOp(alloc=op.target).name
            if ofl_op_name in occurrences:
                ofl_op = op_list[max(occurrences[ofl_op_name])]
                op.wait_events.append((ofl_op.op_type, ofl_op.target.name))
    return op_list

def group(md: ModelPULPOffload, w, tol=1):
    # Group the parameters of each block for the task
    fwd_i = min(md.param2hcn[w])
    bwd_i = max(md.param2hcn[w])
    early_fwd = []
    for t in range(bwd_i, md.T):
        if not md.single_fwd and md.sol(md.sumComp[t, fwd_i]):
            early_fwd.append(t)  # if recompute fwd after bwd
    hcn = md.hgraph.list_HCNs[fwd_i]
    parameters = {pnode.param_name: pnode for pnode in md.parameters[w]}
    parameter_size = sum(pnode.mem for pnode in parameters.values())

    Alive = {p: 1 for p in parameters.keys()}
    Offloaded = {p: False for p in parameters.keys()}

    ofl_ops = []
    prf_ops = []
    del_ops = []
    opt_ops = []
    init_ops = []
    restore_ops = []
    init_alive = []
    cpu_optimize_candidates = {p: 0 for p in parameters.keys()} if md.optimize_metrics else {}

    def apply_cpu_optimize(p):
        for t, k, op in ofl_ops:
            if op.target.target_name == p:
                op.target.is_grad = True
                # op.grad = True
                break
        # for (t,k,op) in del_ops:
        #     if op.target.name == p:
        #         op.grad = True
        del_ops.append((t, k, DeleteOp(Parameter(parameters[p], is_grad=True))))
        i = (
            md.active_steps.index((t, k)) + 1
        )  # TODO: distribute cpu optimization based on time
        p_alloc = Parameter(parameters[p])
        op = OptimizeOp(
            list_params=[p],
            cpu=True,
            alloc=p_alloc,
            time=parameters[p].mem / md.cpu_optimize_speed / md.gcd,
        )
        opt_ops.append((*md.active_steps[i], op))
        md.cpu_optimized_steps[md.active_steps[i]].append(p)
        # del_ops.append((bwd_i, bwd_i, DeleteOp(Parameter(parameters[p]))))

        # if cpu optimize, do not keep w after bwd

    def apply_gpu_optimize(p):
        p_alloc = Parameter(parameters[p])
        op = OptimizeOp(
            list_params=[p],
            alloc=p_alloc,
            time=parameters[p].mem / md.gpu_optimize_speed / md.gcd,
            overhead=parameters[p].mem * md.optimizer_overhead_factor,
        )
        opt_ops.append((bwd_i, bwd_i, op))  # optimize after bwd
        del_ops.append((bwd_i, bwd_i, DeleteOp(Parameter(parameters[p], is_grad=True))))

    assert (bwd_i, bwd_i) in md.active_steps
    idx = md.active_steps.index((bwd_i, bwd_i))
    for t, k in md.active_steps[idx:] + md.active_steps[:idx]:
        t_, k_ = md.next_idx(t, k)
        current_alive_size = sum(parameters[p].mem * a for p, a in Alive.items())
        current_offloaded_size = sum(
            parameters[p].mem * a for p, a in Offloaded.items()
        )
        next_alive_size = round(
            (md.AliveG[(t_, k_, w)] + md.AliveW[(t_, k_, w)]).value() * parameter_size
        )
        next_offloaded_size = round(
            (md.OflGProg[(t_, k_, w)] + md.OflWProg[(t_, k_, w)]).value()
            * parameter_size
        )

        # assert current_alive_size <= round(md.AliveW[(t, k, w)].value() * parameter_size)

        if (t, k) == (0, 0):  # init
            for p, a in Alive.items():
                if a:
                    p_alloc = Parameter(parameters[p])
                    # init_ops.append((t, k, AllocateOp(p_alloc)))
                    init_ops.append(
                        (t, k, PrepareOp(p_alloc, device="cuda", cpu_optimize=True))
                    )
                    init_alive.append(p_alloc)
                    # op = PrefetchOp(
                    #     alloc=p_alloc, indices=(0, None),
                    #     time=parameters[p].mem/md.bandwidthPrf/md.gcd
                    # )
                    # init_ops.append((t, k, op))
                    op = OffloadOp(
                        alloc=p_alloc,
                        indices=(0, None),
                        time=parameters[p].mem / md.bandwidthOfl / md.gcd,
                    )
                    restore_ops.append((t, k, op))
                    restore_ops.append((t, k, DeleteOp(p_alloc)))

        if next_offloaded_size > current_offloaded_size:
            # print(t,k, next_offloaded_size, current_offloaded_size)
            ofl_size = next_offloaded_size - current_offloaded_size
            candidates = {
                p: parameters[p].mem * (1 - o) for p, o in Offloaded.items() if o < 1
            }
            if not candidates:
                if ofl_size < 1024:
                    ofl_size = 0
                else:
                    raise ValueError
            selector = knapsack(list(candidates.items()))
            select_paras = selector.select_size(ofl_size)
            # assert ofl_size==0 or sum(candidates[p] for p in select_paras)/ofl_size>0.99
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-ofl_size>tol:
            #     pass
            for p in select_paras:
                op = OffloadOp(
                    alloc=Parameter(parameters[p]),
                    indices=(0, None),
                    time=parameters[p].mem / md.bandwidthOfl / md.gcd,
                )
                ofl_ops.append((t, k, op))
                Offloaded[p] = 1

        if current_alive_size > next_alive_size:
            del_size = current_alive_size - next_alive_size
            candidates = {}
            for p, o in Offloaded.items():
                if Alive[p] > 0 and o > 0:
                    candidates[p] = min(o, Alive[p]) * parameters[p].mem
            if not candidates:
                if del_size < 1024:
                    del_size = 0
                else:
                    raise ValueError
            selector = knapsack(list(candidates.items()))
            select_paras = selector.select_size(del_size)
            # assert del_size==0 or sum(candidates[p] for p in select_paras)/del_size>0.99
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-del_size>tol:
            #     pass
            for p in select_paras:
                del_ops.append((t, k, DeleteOp(Parameter(parameters[p]))))
                Alive[p] = 0
        if current_alive_size < next_alive_size:
            # prefetch should be smaller than solution
            prf_size = next_alive_size - current_alive_size
            candidates = {
                p: parameters[p].mem * (1 - a) for p, a in Alive.items() if a < 1
            }
            if not candidates:
                if prf_size < 1024:
                    prf_size = 0
                else:
                    raise ValueError
            if md.sol(md.AliveW[(t_, k_, w)]):
                select_paras = list(candidates.keys())
                # assert prf_size==0 or sum(candidates[p] for p in select_paras)/prf_size>0.99
            else:
                selector = knapsack(list(candidates.items()))
                unselect_paras = selector.select_size(
                    sum(candidates.values()) - prf_size
                )

                select_paras = [p for p in candidates.keys() if p not in unselect_paras]
                # assert prf_size==0 or sum(candidates[p] for p in select_paras)/prf_size<1.01
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-prf_size>tol:
            #     pass
            for p in select_paras:
                prf_ops.append((t, k, AllocateOp(Parameter(parameters[p]))))
                op = PrefetchOp(
                    alloc=Parameter(parameters[p]),
                    indices=(0, None),
                    time=parameters[p].mem / md.bandwidthPrf / md.gcd,
                )
                prf_ops.append((t, k, op))
                Alive[p] = 1
                if (t > bwd_i and t < min(early_fwd + [md.T + 1])) or t < fwd_i:
                    # cpu optimize only if prefetch before fwd
                    if parameters[p].info.requires_grad and md.optimize_metrics:
                        # only trainable parameters will be optimize candidate
                        cpu_optimize_candidates[p] = 1

    if not md.optimize_metrics:
        return ofl_ops, prf_ops, del_ops, opt_ops, init_ops, restore_ops, init_alive

    candidates = {
        p: parameters[p].mem * a for p, a in cpu_optimize_candidates.items() if a > 0
    }
    select_paras = []
    # assert sum(candidates.values())/parameter_size >= md.sumOptC[w].value()-0.01

    # cpu_optimize_size = md.sumOptC[w].value()*parameter_size# size by subgraph
    if isinstance(md.req_w, float):
        multiplier = 1 - md.req_w
    else:
        multiplier = 1 - md.req_w.value()
    # cpu_optimize_size = (sum(md.sumOptC[w_].value() *
    #                         md.parameter_gradient_size[w_] *md.gcd
    #                         for w_ in range(w, md.W)) / (1-multiplier)
    #                     - sum(md.cpu_optimized_params.values()))# size by all graphs
    cpu_optimize_size = min(
        sum(candidates.values()),
        (md.sumOptC[w].value() * md.parameter_gradient_size[w] * md.gcd)
        / (1 - multiplier),
    )

    if candidates and cpu_optimize_size > 0:
        # print(candidates, cpu_optimize_size)
        selector = knapsack(list(candidates.items()))
        select_paras = selector.select_size(cpu_optimize_size)
        if cpu_optimize_size > sum(candidates[p] for p in select_paras):
            raise ValueError
        # print(select_paras)

    # Optimize parameters which requires grad
    gpu_optimze_param = []
    for p, pnode in parameters.items():
        if not pnode.info.requires_grad:
            continue
        if p in select_paras:
            md.cpu_optimized_params[p] = parameters[p].mem
            apply_cpu_optimize(p)
        else:
            apply_gpu_optimize(p)
            gpu_optimze_param.append(pnode)
    if md.optimize_metrics and gpu_optimze_param:
        ofl_ops_os, prf_ops_os, del_ops_os, init_alive_os = group_optimizer_states(
            md, w, gpu_optimze_param
        )
        ofl_ops += ofl_ops_os
        prf_ops += prf_ops_os
        del_ops += del_ops_os
        init_alive += init_alive_os
    return ofl_ops, prf_ops, del_ops, opt_ops, init_ops, restore_ops, init_alive


def group_optimizer_states(md, w, gpu_optimize_param):
    # To offload and prefetch optimizer states witin the gpu_optimize_param
    ofl_ops = []
    prf_ops = []
    del_ops = []
    init_alive = []
    fwd_i = min(md.param2hcn[w])
    bwd_i = max(md.param2hcn[w])
    hcn = md.hgraph.list_HCNs[fwd_i]
    parameters = {
        pnode.param_name: pnode for pnode in md.parameters[w] if pnode.requires_grad
    }
    parameter_size = sum(pnode.mem for pnode in parameters.values())
    gpu_optimize_size = sum(pnode.mem for pnode in gpu_optimize_param)

    Alive = {pnode.param_name: 1 for pnode in gpu_optimize_param}
    Offloaded = {pnode.param_name: False for pnode in gpu_optimize_param}
    assert (bwd_i, bwd_i) in md.active_steps
    idx = md.active_steps.index((bwd_i, bwd_i))
    for t, k in md.active_steps[idx:] + md.active_steps[:idx]:
        if (t, k) == (0, 0):  # init
            for p, a in Alive.items():
                if a:
                    init_alive.append(Parameter(parameters[p], is_optim_states=True))

        t_, k_ = md.next_idx(t, k)
        current_alive_size = sum(parameters[p].mem * a for p, a in Alive.items())
        current_offloaded_size = sum(
            parameters[p].mem * a for p, a in Offloaded.items()
        )
        next_alive_size = min(
            gpu_optimize_size, round((md.AliveO[(t_, k_, w)]).value() * parameter_size)
        )
        next_offloaded_size = min(
            gpu_optimize_size,
            round((md.OflOProg[(t_, k_, w)]).value() * parameter_size),
        )
        if parameter_size * (1 - md.sumOptC[w]).value() < gpu_optimize_size:
            next_offloaded_size += (
                gpu_optimize_size - parameter_size * (1 - md.sumOptC[w]).value()
            )

        # assert current_alive_size <= round(md.AliveW[(t, k, w)].value() * parameter_size)
        if next_offloaded_size > current_offloaded_size:
            # print(t,k, next_offloaded_size, current_offloaded_size)
            ofl_size = next_offloaded_size - current_offloaded_size
            candidates = {
                p: parameters[p].mem * (1 - o) for p, o in Offloaded.items() if o < 1
            }
            if not candidates:
                if ofl_size < 1024:
                    ofl_size = 0
                else:
                    raise ValueError
            selector = knapsack(list(candidates.items()))
            select_paras = selector.select_size(ofl_size)
            # assert ofl_size==0 or sum(candidates[p] for p in select_paras)/ofl_size>0.99
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-ofl_size>tol:
            #     pass
            for p in select_paras:
                op = OffloadOp(
                    alloc=Parameter(parameters[p], is_optim_states=True),
                    indices=(0, None),
                    time=parameters[p].mem
                    / md.bandwidthOfl
                    / md.gcd
                    * md.optimizer_states_factor,
                )
                ofl_ops.append((t, k, op))
                Offloaded[p] = 1

        if current_alive_size > next_alive_size:
            if k_ == bwd_i:
                continue
            del_size = current_alive_size - next_alive_size
            candidates = {}
            for p, o in Offloaded.items():
                if Alive[p] > 0 and o > 0:
                    candidates[p] = min(o, Alive[p]) * parameters[p].mem
            if not candidates:
                if del_size < 1024:
                    del_size = 0
                else:
                    raise ValueError
            selector = knapsack(list(candidates.items()))
            select_paras = selector.select_size(del_size)
            # assert del_size==0 or sum(candidates[p] for p in select_paras)/del_size>0.99
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-del_size>tol:
            #     pass
            for p in select_paras:
                del_ops.append(
                    (t, k, DeleteOp(Parameter(parameters[p], is_optim_states=True)))
                )
                Alive[p] = 0
        if current_alive_size < next_alive_size or k_ == bwd_i:
            # if w == 15:print(md.active_steps[k_]==bwd_i)
            # prefetch should be smaller than solution
            prf_size = next_alive_size - current_alive_size
            candidates = {
                p: parameters[p].mem * (1 - a) for p, a in Alive.items() if a < 1
            }
            if not candidates:
                if prf_size < 1024:
                    prf_size = 0
                else:
                    raise ValueError
            if (
                md.sol(md.AliveO[(t_, k_, w)] + md.sumOptC[w] - md.req_w)
                or k_ == bwd_i
            ):

                select_paras = list(candidates.keys())
                # assert prf_size==0 or sum(candidates[p] for p in select_paras)/prf_size>0.99
            else:
                selector = knapsack(list(candidates.items()))
                unselect_paras = selector.select_size(
                    sum(candidates.values()) - prf_size
                )

                select_paras = [p for p in candidates.keys() if p not in unselect_paras]
                # assert prf_size==0 or sum(candidates[p] for p in select_paras)/prf_size<1.01
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-prf_size>tol:
            #     pass
            for p in select_paras:
                prf_ops.append(
                    (t, k, AllocateOp(Parameter(parameters[p], is_optim_states=True)))
                )
                op = PrefetchOp(
                    alloc=Parameter(parameters[p], is_optim_states=True),
                    indices=(0, None),
                    time=parameters[p].mem
                    / md.bandwidthPrf
                    / md.gcd
                    * md.optimizer_states_factor,
                )
                prf_ops.append((t, k, op))
                Alive[p] = 1
        if k_ == bwd_i:
            assert 0 not in Alive.values()

    return ofl_ops, prf_ops, del_ops, init_alive


def group_activation_offload(md: ModelPULPOffload, j):
    ofl_ops = []
    prf_ops = []
    del_ops = []

    if (
        not md.activation_offload or not md.single_fwd
    ):  # we assume activation is based on single fwd
        return ofl_ops, prf_ops, del_ops

    fwd_i = min(md.sub_c2hcn[j])
    bwd_i = max(md.sub_c2hcn[j])
    hcn = md.hgraph.list_HCNs[fwd_i]
    sub_cluster = hcn.sub_cluster

    if md.sol(md.sumComp[fwd_i, fwd_i]):
        opt = -1
        for o in range(md.nSched[fwd_i]):
            if md.sol(md.Comp[fwd_i, fwd_i, o]):
                opt = o
                break

    assert opt > -1

    # phantoms = {}
    # for re_anode in md.list_list_sched[j][opt].phantoms:
    #     anode = sub_cluster.translate_representee_node(re_anode)
    #     phantoms[anode.name] = anode
    phantoms = {anode.name: anode for anode in md.phantoms[j, opt]}
    md.selected_phantoms[j] = [anode.main_target for anode in phantoms.values()]
    phantom_size = sum(anode.mem for anode in phantoms.values())

    Alive = {anode.name: 1 for anode in phantoms.values()}
    # Offloaded = {anode.name: False for anode in phantoms.values()}
    assert (bwd_i, bwd_i) in md.active_steps

    idx_f = md.active_steps.index((fwd_i, fwd_i))
    idx_b = md.active_steps.index((bwd_i, bwd_i))
    for t, k in md.active_steps[idx_f:] + md.active_steps[:idx_b]:
        t_, k_ = md.next_idx(t, k)

        # current_alive_size = sum(phantoms[n].mem * a for n, a in Alive.items())
        current_alive_size = sum(phantoms[n].mem * a for n, a in Alive.items() if a > 0)

        next_alive_size = phantom_size - md.OflPProg[t, k, j].value() * md.gcd

        # next_offloaded_size = min(gpu_optimize_size,
        #     round((md.OflOProg[(t_, k_, w)]).value() * parameter_size))
        # if parameter_size * (1-md.sumOptC[w]).value()<gpu_optimize_size:
        #     next_offloaded_size += gpu_optimize_size - parameter_size * (1-md.sumOptC[w]).value()

        # assert current_alive_size <= round(md.AliveW[(t, k, w)].value() * parameter_size)
        if current_alive_size > next_alive_size:
            # print(t,k, next_offloaded_size, current_offloaded_size)
            ofl_size = current_alive_size - next_alive_size
            candidates = {n: phantoms[n].mem * a for n, a in Alive.items() if a > 0}
            if not candidates:
                if ofl_size < 1024:
                    ofl_size = 0
                else:
                    raise ValueError
            selector = knapsack(list(candidates.items()))
            select_paras = selector.select_size(ofl_size)
            # assert ofl_size==0 or sum(candidates[p] for p in select_paras)/ofl_size>0.99
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-ofl_size>tol:
            #     pass
            for n in select_paras:
                op = OffloadOp(
                    alloc=Activation(phantoms[n]),
                    indices=(0, None),
                    time=phantoms[n].mem / md.bandwidthOfl / md.gcd,
                )
                ofl_ops.append((t, k, op))
                del_ops.append((t, k, DeleteOp(Activation(phantoms[n]))))
                Alive[n] = 0

        # if current_alive_size > next_alive_size:
        #     if k_ ==bwd_i:continue
        #     del_size = current_alive_size - next_alive_size
        #     candidates = {}
        #     for p, o in Offloaded.items():
        #         if Alive[p]>0 and o>0:
        #             candidates[p] = min(o, Alive[p])*parameters[p].mem
        #     if not candidates:
        #         if del_size<1024:
        #             del_size = 0
        #         else:
        #             raise ValueError
        #     selector = knapsack(list(candidates.items()))
        #     select_paras = selector.select_size(del_size)
        #     # assert del_size==0 or sum(candidates[p] for p in select_paras)/del_size>0.99
        #     # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-del_size>tol:
        #     #     pass
        #     for p in select_paras:
        #         del_ops.append((t, k, DeleteOp(Parameter(parameters[p],
        #                                                     is_optim_states=True)
        #                                         )))
        #         Alive[p] = 0
        if current_alive_size < next_alive_size or k_ == bwd_i:
            # if w == 15:print(md.active_steps[k_]==bwd_i)
            # prefetch should be smaller than solution
            prf_size = next_alive_size - current_alive_size
            candidates = {
                n: phantoms[n].mem * (1 - a) for n, a in Alive.items() if a < 1
            }
            if not candidates:
                if prf_size < 1024:
                    prf_size = 0
                else:
                    raise ValueError
            if k_ == bwd_i:
                select_paras = list(candidates.keys())
                # assert prf_size==0 or sum(candidates[p] for p in select_paras)/prf_size>0.99
            else:
                selector = knapsack(list(candidates.items()))
                unselect_paras = selector.select_size(
                    sum(candidates.values()) - prf_size
                )

                select_paras = [p for p in candidates.keys() if p not in unselect_paras]
                # assert prf_size==0 or sum(candidates[p] for p in select_paras)/prf_size<1.01
            # if sum(candidates[p] for p in select_paras)/sum(candidates.values())-prf_size>tol:
            #     pass
            for n in select_paras:
                alloc = Activation(phantoms[n])
                prf_ops.append((t, k, AllocateOp(alloc)))
                op = PrefetchOp(
                    alloc=alloc,
                    indices=(0, None),
                    time=alloc.mem / md.bandwidthPrf / md.gcd,
                )
                prf_ops.append((t, k, op))
                Alive[n] = 1
        if k_ == bwd_i:
            assert 0 not in Alive.values()

    return ofl_ops, prf_ops, del_ops
